from datetime import datetime, timezone
from functools import lru_cache
from typing import TYPE_CHECKING, Dict, NamedTuple, Optional, cast

from pydantic import Field
from pydantic.types import NonNegativeInt

from datahub.ingestion.api.ingestion_job_checkpointing_provider_base import JobId
from datahub.ingestion.source.state.checkpoint import Checkpoint, CheckpointStateBase
from datahub.ingestion.source.state.use_case_handler import (
    StatefulIngestionUsecaseHandlerBase,
)

if TYPE_CHECKING:
    from datahub.ingestion.source.datahub.datahub_source import DataHubSource


class DataHubIngestionState(CheckpointStateBase):
    database_createdon_ts: NonNegativeInt = 0

    # Maps partition -> offset
    kafka_offsets: Dict[int, NonNegativeInt] = Field(default_factory=dict)

    @property
    def database_createdon_datetime(self) -> datetime:
        return datetime.fromtimestamp(
            self.database_createdon_ts / 1000, tz=timezone.utc
        )


class PartitionOffset(NamedTuple):
    partition: int
    offset: int


class StatefulDataHubIngestionHandler(
    StatefulIngestionUsecaseHandlerBase[DataHubIngestionState]
):
    def __init__(self, source: "DataHubSource"):
        self.state_provider = source.state_provider
        self.config = source.config.stateful_ingestion
        self.run_id = source.ctx.run_id
        self.pipeline_name = source.ctx.pipeline_name
        self.state_provider.register_stateful_ingestion_usecase_handler(self)

    @lru_cache(maxsize=1)
    def is_checkpointing_enabled(self) -> bool:
        return self.state_provider.is_stateful_ingestion_configured()

    def get_last_run_state(self) -> DataHubIngestionState:
        if self.is_checkpointing_enabled() and not self.config.ignore_old_state:
            last_checkpoint = self.state_provider.get_last_checkpoint(
                self.job_id, DataHubIngestionState
            )
            if last_checkpoint and last_checkpoint.state:
                return last_checkpoint.state

        return DataHubIngestionState()

    def create_checkpoint(self) -> Optional[Checkpoint[DataHubIngestionState]]:
        if not self.is_checkpointing_enabled() or self.config.ignore_new_state:
            return None

        if self.pipeline_name is None:
            raise ValueError(
                "Pipeline name must be set to use stateful datahub ingestion"
            )

        return Checkpoint(
            job_name=self.job_id,
            pipeline_name=self.pipeline_name,
            run_id=self.run_id,
            state=self.get_last_run_state(),
        )

    def update_checkpoint(
        self,
        *,
        last_createdon: Optional[datetime] = None,
        last_offset: Optional[PartitionOffset] = None,
    ) -> None:
        cur_checkpoint = self.state_provider.get_current_checkpoint(self.job_id)
        if cur_checkpoint:
            cur_state = cast(DataHubIngestionState, cur_checkpoint.state)
            if last_createdon:
                cur_state.database_createdon_ts = int(last_createdon.timestamp() * 1000)
            if last_offset:
                cur_state.kafka_offsets[last_offset.partition] = last_offset.offset + 1

    def commit_checkpoint(self) -> None:
        if self.state_provider.ingestion_checkpointing_state_provider:
            self.state_provider.prepare_for_commit()
            self.state_provider.ingestion_checkpointing_state_provider.commit()

    @property
    def job_id(self) -> JobId:
        return JobId("datahub_ingestion")



import logging
from typing import Iterable, List

from datahub.emitter.mce_builder import make_dataset_urn_with_platform_instance
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.snowflake.snowflake_config import (
    DatabaseId,
    SnowflakeV2Config,
)
from datahub.ingestion.source.snowflake.snowflake_report import SnowflakeV2Report
from datahub.ingestion.source.snowflake.snowflake_schema import SnowflakeDatabase
from datahub.ingestion.source.snowflake.snowflake_utils import SnowflakeCommonMixin
from datahub.metadata.com.linkedin.pegasus2avro.common import Siblings
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    DatasetLineageType,
    Upstream,
    UpstreamLineage,
)

logger: logging.Logger = logging.getLogger(__name__)


class SnowflakeSharesHandler(SnowflakeCommonMixin):
    def __init__(
        self,
        config: SnowflakeV2Config,
        report: SnowflakeV2Report,
    ) -> None:
        self.config = config
        self.report = report

    def get_shares_workunits(
        self, databases: List[SnowflakeDatabase]
    ) -> Iterable[MetadataWorkUnit]:
        inbounds = self.config.inbounds()
        outbounds = self.config.outbounds()
        # None of the databases are shared
        if not (inbounds or outbounds):
            return

        logger.debug("Checking databases for inbound or outbound shares.")
        for db in databases:
            is_inbound = db.name in inbounds
            is_outbound = db.name in outbounds

            if not (is_inbound or is_outbound):
                logger.debug(f"database {db.name} is not shared.")
                continue

            sibling_dbs = (
                list(outbounds[db.name]) if is_outbound else [inbounds[db.name]]
            )

            for schema in db.schemas:
                for table_name in schema.tables + schema.views:
                    # TODO: If this is outbound database,
                    # 1. attempt listing shares using `show shares` to identify name of share associated with this database (cache query result).
                    # 2. if corresponding share is listed, then run `show grants to share <share_name>` to identify exact tables, views included in share.
                    # 3. emit siblings only for the objects listed above.
                    # This will work only if the configured role has accountadmin role access OR is owner of share.
                    # Otherwise ghost nodes may be shown in "Composed Of" section for tables/views in original database which are not granted to share.
                    yield from self.gen_siblings(
                        db.name,
                        schema.name,
                        table_name,
                        is_outbound,
                        sibling_dbs,
                    )

                    if is_inbound:
                        assert len(sibling_dbs) == 1
                        # SnowflakeLineageExtractor is unaware of database->schema->table hierarchy
                        # hence this lineage code is not written in SnowflakeLineageExtractor
                        # also this is not governed by configs include_table_lineage and include_view_lineage
                        yield self.get_upstream_lineage_with_primary_sibling(
                            db.name, schema.name, table_name, sibling_dbs[0]
                        )

        self.report_missing_databases(
            databases, list(inbounds.keys()), list(outbounds.keys())
        )

    def report_missing_databases(
        self,
        databases: List[SnowflakeDatabase],
        inbounds: List[str],
        outbounds: List[str],
    ) -> None:
        db_names = [db.name for db in databases]
        missing_dbs = [db for db in inbounds + outbounds if db not in db_names]

        if missing_dbs and self.config.platform_instance:
            self.report.warning(
                title="Extra Snowflake share configurations",
                message="Some databases referenced by the share configs were not ingested. Siblings/lineage will not be set for these.",
                context=f"{missing_dbs}",
            )
        elif missing_dbs:
            logger.debug(
                f"Databases {missing_dbs} were not ingested in this recipe.",
            )

    def gen_siblings(
        self,
        database_name: str,
        schema_name: str,
        table_name: str,
        primary: bool,
        sibling_databases: List[DatabaseId],
    ) -> Iterable[MetadataWorkUnit]:
        if not sibling_databases:
            return
        dataset_identifier = self.identifiers.get_dataset_identifier(
            table_name, schema_name, database_name
        )
        urn = self.identifiers.gen_dataset_urn(dataset_identifier)

        sibling_urns = [
            make_dataset_urn_with_platform_instance(
                self.identifiers.platform,
                self.identifiers.get_dataset_identifier(
                    table_name, schema_name, sibling_db.database
                ),
                sibling_db.platform_instance,
            )
            for sibling_db in sibling_databases
        ]

        yield MetadataChangeProposalWrapper(
            entityUrn=urn,
            aspect=Siblings(primary=primary, siblings=sorted(sibling_urns)),
        ).as_workunit()

    def get_upstream_lineage_with_primary_sibling(
        self,
        database_name: str,
        schema_name: str,
        table_name: str,
        primary_sibling_db: DatabaseId,
    ) -> MetadataWorkUnit:
        dataset_identifier = self.identifiers.get_dataset_identifier(
            table_name, schema_name, database_name
        )
        urn = self.identifiers.gen_dataset_urn(dataset_identifier)

        upstream_urn = make_dataset_urn_with_platform_instance(
            self.identifiers.platform,
            self.identifiers.get_dataset_identifier(
                table_name, schema_name, primary_sibling_db.database
            ),
            primary_sibling_db.platform_instance,
        )

        return MetadataChangeProposalWrapper(
            entityUrn=urn,
            aspect=UpstreamLineage(
                upstreams=[Upstream(dataset=upstream_urn, type=DatasetLineageType.COPY)]
            ),
        ).as_workunit()



import logging
from dataclasses import dataclass
from typing import List, Tuple

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

_LOOKAROUND_LINES = 300

# The Python difflib library can generate unified diffs, but it cannot apply them.
# There weren't any well-maintained and easy-to-use libraries for applying
# unified diffs, so I wrote my own.
#
# My implementation is focused on ensuring correctness, and will throw
# an exception whenever it detects an issue.
#
# Alternatives considered:
# - diff-match-patch: This was the most promising since it's from Google.
#       Unfortunately, they deprecated the library in Aug 2024. That may not have
#       been a dealbreaker, since a somewhat greenfield community fork exists:
#       https://github.com/dmsnell/diff-match-patch
#       However, there's also a long-standing bug in the library around the
#       handling of line breaks when parsing diffs. See:
#       https://github.com/google/diff-match-patch/issues/157
# - python-patch: Seems abandoned.
# - patch-ng: Fork of python-patch, but mainly targeted at applying patches to trees.
#       It did not have simple "apply patch to string" abstractions.
# - unidiff: Parses diffs, but cannot apply them.


class InvalidDiffError(Exception):
    pass


class DiffApplyError(Exception):
    pass


@dataclass
class Hunk:
    source_start: int
    source_lines: int
    target_start: int
    target_lines: int
    lines: List[Tuple[str, str]]


def parse_patch(patch_text: str) -> List[Hunk]:
    """
    Parses a unified diff patch into a list of Hunk objects.

    Args:
        patch_text: Unified diff format patch text

    Returns:
        List of parsed Hunk objects

    Raises:
        InvalidDiffError: If the patch is in an invalid format
    """
    hunks = []
    patch_lines = patch_text.splitlines()
    i = 0

    while i < len(patch_lines):
        line = patch_lines[i]

        if line.startswith("@@"):
            try:
                header_parts = line.split()
                if len(header_parts) < 3:
                    raise ValueError(f"Invalid hunk header format: {line}")

                source_changes, target_changes = header_parts[1:3]
                source_start, source_lines = map(int, source_changes[1:].split(","))
                target_start, target_lines = map(int, target_changes[1:].split(","))

                hunk = Hunk(source_start, source_lines, target_start, target_lines, [])
                i += 1

                while i < len(patch_lines) and not patch_lines[i].startswith("@@"):
                    hunk_line = patch_lines[i]
                    if hunk_line:
                        hunk.lines.append((hunk_line[0], hunk_line[1:]))
                    else:
                        # Fully empty lines usually means an empty context line that was
                        # trimmed by trailing whitespace removal.
                        hunk.lines.append((" ", ""))
                    i += 1

                hunks.append(hunk)
            except (IndexError, ValueError) as e:
                raise InvalidDiffError(f"Failed to parse hunk: {str(e)}") from e
        else:
            raise InvalidDiffError(f"Invalid line format: {line}")

    return hunks


def find_hunk_start(source_lines: List[str], hunk: Hunk) -> int:
    """
    Finds the actual starting line of a hunk in the source lines.

    Args:
        source_lines: The original source lines
        hunk: The hunk to locate

    Returns:
        The actual line number where the hunk starts

    Raises:
        DiffApplyError: If the hunk's context cannot be found in the source lines
    """

    # Extract context lines from the hunk, stopping at the first non-context line
    context_lines = []
    for prefix, line in hunk.lines:
        if prefix == " ":
            context_lines.append(line)
        else:
            break

    if not context_lines:
        logger.debug("No context lines found in hunk.")
        return hunk.source_start - 1  # Default to the original start if no context

    logger.debug(
        f"Searching for {len(context_lines)} context lines, starting with {context_lines[0]}"
    )

    # Define the range to search for the context lines
    search_start = max(0, hunk.source_start - _LOOKAROUND_LINES)
    search_end = min(len(source_lines), hunk.source_start + _LOOKAROUND_LINES)

    # Iterate over the possible starting positions in the source lines
    for i in range(search_start, search_end):
        # Check if the context lines match the source lines starting at position i
        match = True
        for j, context_line in enumerate(context_lines):
            if (i + j >= len(source_lines)) or source_lines[i + j] != context_line:
                match = False
                break
        if match:
            # logger.debug(f"Context match found at line: {i}")
            return i

    logger.debug(f"Could not find match for hunk context lines: {context_lines}")
    raise DiffApplyError("Could not find match for hunk context.")


def apply_hunk(result_lines: List[str], hunk: Hunk, hunk_index: int) -> None:
    """
    Applies a single hunk to the result lines.

    Args:
        result_lines: The current state of the patched file
        hunk: The hunk to apply
        hunk_index: The index of the hunk (for logging purposes)

    Raises:
        DiffApplyError: If the hunk cannot be applied correctly
    """
    current_line = find_hunk_start(result_lines, hunk)
    logger.debug(f"Hunk {hunk_index + 1} start line: {current_line}")

    for line_index, (prefix, content) in enumerate(hunk.lines):
        # logger.debug(f"Processing line {line_index + 1} of hunk {hunk_index + 1}")
        # logger.debug(f"Current line: {current_line}, Total lines: {len(result_lines)}")
        # logger.debug(f"Prefix: {prefix}, Content: {content}")

        if current_line >= len(result_lines):
            logger.debug(f"Reached end of file while applying hunk {hunk_index + 1}")
            while line_index < len(hunk.lines) and hunk.lines[line_index][0] == "+":
                result_lines.append(hunk.lines[line_index][1])
                line_index += 1

            # If there's context or deletions past the end of the file, that's an error.
            if line_index < len(hunk.lines):
                raise DiffApplyError(
                    f"Found context or deletions after end of file in hunk {hunk_index + 1}"
                )
            break

        if prefix == "-":
            if result_lines[current_line].strip() != content.strip():
                raise DiffApplyError(
                    f"Removing line that doesn't exactly match. Expected: '{content.strip()}', Found: '{result_lines[current_line].strip()}'"
                )
            result_lines.pop(current_line)
        elif prefix == "+":
            result_lines.insert(current_line, content)
            current_line += 1
        elif prefix == " ":
            if result_lines[current_line].strip() != content.strip():
                raise DiffApplyError(
                    f"Context line doesn't exactly match. Expected: '{content.strip()}', Found: '{result_lines[current_line].strip()}'"
                )
            current_line += 1
        else:
            raise DiffApplyError(
                f"Invalid line prefix '{prefix}' in hunk {hunk_index + 1}, line {line_index + 1}"
            )


def apply_diff(source: str, patch_text: str) -> str:
    """
    Applies a unified diff patch to source text and returns the patched result.

    Args:
        source: Original source text to be patched
        patch_text: Unified diff format patch text (with @@ markers and hunks)

    Returns:
        The patched text result

    Raises:
        InvalidDiffError: If the patch is in an invalid format
        DiffApplyError: If the patch cannot be applied correctly
    """

    # logger.debug(f"Original source:\n{source}")
    # logger.debug(f"Patch text:\n{patch_text}")

    hunks = parse_patch(patch_text)
    logger.debug(f"Parsed into {len(hunks)} hunks")

    source_lines = source.splitlines()
    result_lines = source_lines.copy()

    for hunk_index, hunk in enumerate(hunks):
        logger.debug(f"Processing hunk {hunk_index + 1}")
        apply_hunk(result_lines, hunk, hunk_index)

    result = "\n".join(result_lines) + "\n"
    # logger.debug(f"Patched result:\n{result}")
    return result



import logging
from typing import List, Optional, Tuple

from lark import Tree

from datahub.ingestion.source.powerbi.m_query import resolver, tree_function

logger = logging.getLogger(__name__)


def validate_parse_tree(
    tree: Tree, native_query_enabled: bool = True
) -> Tuple[bool, Optional[str]]:
    """
    :param tree: tree to validate as per functions supported by m_parser module
    :param native_query_enabled: Whether user want to extract lineage from native query
    :return: first argument is False if validation is failed and second argument would contain the error message.
             in the case of valid tree, the first argument is True and the second argument would be None.
    """
    functions: List[str] = tree_function.get_all_function_name(tree)
    if len(functions) == 0:
        return False, "Function calls not found"

    if native_query_enabled is False:
        if resolver.FunctionName.NATIVE_QUERY.value in functions:
            return False, "Lineage extraction from native query is disabled."

    return True, None



from dataclasses import dataclass
from datetime import datetime

from datahub.ingestion.source.sql.sql_generic_profiler import ProfilingSqlReport
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
)
from datahub.ingestion.source_report.ingestion_stage import IngestionStageReport


@dataclass
class DremioSourceReport(
    ProfilingSqlReport, StaleEntityRemovalSourceReport, IngestionStageReport
):
    num_containers_failed: int = 0
    num_datasets_failed: int = 0

    def report_upstream_latency(self, start_time: datetime, end_time: datetime) -> None:
        # recording total combined latency is not very useful, keeping this method as a placeholder
        # for future implementation of min / max / percentiles etc.
        pass

    def report_entity_scanned(self, name: str, ent_type: str = "View") -> None:
        """
        Entity could be a view or a table
        """
        if ent_type == "Table":
            self.tables_scanned += 1
        elif ent_type == "View":
            self.views_scanned += 1
        else:
            raise KeyError(f"Unknown entity {ent_type}.")

    def set_ingestion_stage(self, dataset: str, stage: str) -> None:
        self.report_ingestion_stage_start(f"{dataset}: {stage}")



import datetime
import json
import logging
from json import JSONDecodeError
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    MutableMapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Union,
)

from looker_sdk.error import SDKError
from looker_sdk.rtl.serialize import DeserializeError
from looker_sdk.sdk.api40.models import (
    Dashboard,
    DashboardElement,
    Folder,
    FolderBase,
    Look,
    LookWithQuery,
    Query,
)

import datahub.emitter.mce_builder as builder
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import create_embed_mcp, gen_containers
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    MetadataWorkUnitProcessor,
    SourceCapability,
    SourceReport,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.source_helpers import auto_workunit
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import (
    BIAssetSubTypes,
    BIContainerSubTypes,
)
from datahub.ingestion.source.looker import looker_usage
from datahub.ingestion.source.looker.looker_common import (
    InputFieldElement,
    LookerDashboard,
    LookerDashboardElement,
    LookerDashboardSourceReport,
    LookerExplore,
    LookerExploreRegistry,
    LookerFolder,
    LookerFolderKey,
    LookerUser,
    LookerUserRegistry,
    LookerUtil,
    ViewField,
    ViewFieldType,
    gen_model_key,
)
from datahub.ingestion.source.looker.looker_config import LookerDashboardSourceConfig
from datahub.ingestion.source.looker.looker_lib_wrapper import LookerAPI
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import (
    AuditStamp,
    ChangeAuditStamps,
    DataPlatformInstance,
    Status,
)
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import (
    ChartSnapshot,
    DashboardSnapshot,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import (
    BrowsePathEntryClass,
    BrowsePathsClass,
    BrowsePathsV2Class,
    ChartInfoClass,
    ChartTypeClass,
    ContainerClass,
    DashboardInfoClass,
    InputFieldClass,
    InputFieldsClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    SubTypesClass,
)
from datahub.utilities.backpressure_aware_executor import BackpressureAwareExecutor

logger = logging.getLogger(__name__)


@platform_name("Looker")
@support_status(SupportStatus.CERTIFIED)
@config_class(LookerDashboardSourceConfig)
@capability(SourceCapability.DESCRIPTIONS, "Enabled by default")
@capability(SourceCapability.PLATFORM_INSTANCE, "Use the `platform_instance` field")
@capability(
    SourceCapability.OWNERSHIP, "Enabled by default, configured using `extract_owners`"
)
@capability(SourceCapability.LINEAGE_COARSE, "Supported by default")
@capability(
    SourceCapability.LINEAGE_FINE,
    "Enabled by default, configured using `extract_column_level_lineage`",
)
@capability(
    SourceCapability.USAGE_STATS,
    "Enabled by default, configured using `extract_usage_history`",
)
class LookerDashboardSource(TestableSource, StatefulIngestionSourceBase):
    """
    This plugin extracts the following:
    - Looker dashboards, dashboard elements (charts) and explores
    - Names, descriptions, URLs, chart types, input explores for the charts
    - Schemas and input views for explores
    - Owners of dashboards

    :::note
    To get complete Looker metadata integration (including Looker views and lineage to the underlying warehouse tables), you must ALSO use the `lookml` module.
    :::
    """

    platform = "looker"

    def __init__(self, config: LookerDashboardSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.source_config: LookerDashboardSourceConfig = config
        self.reporter: LookerDashboardSourceReport = LookerDashboardSourceReport()
        self.looker_api: LookerAPI = LookerAPI(self.source_config)
        self.user_registry: LookerUserRegistry = LookerUserRegistry(self.looker_api)
        self.explore_registry: LookerExploreRegistry = LookerExploreRegistry(
            self.looker_api, self.reporter, self.source_config
        )
        self.reporter._looker_explore_registry = self.explore_registry
        self.reporter._looker_api = self.looker_api

        # Keep track of look-id which are reachable from Dashboard
        self.reachable_look_registry: Set[str] = set()

        # (model, explore) -> list of charts/looks/dashboards that reference this explore
        # The list values are used purely for debugging purposes.
        self.reachable_explores: Dict[Tuple[str, str], List[str]] = {}

        # To keep track of folders (containers) which have already been ingested
        # Required, as we do not ingest all folders but only those that have dashboards/looks
        self.processed_folders: List[str] = []

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        test_report = TestConnectionReport()
        try:
            config = LookerDashboardSourceConfig.parse_obj_allow_extras(config_dict)
            permissions = LookerAPI(config).get_available_permissions()

            test_report.basic_connectivity = CapabilityReport(capable=True)
            test_report.capability_report = {}

            BASIC_INGEST_REQUIRED_PERMISSIONS = {
                # TODO: Make this a bit more granular.
                "access_data",
                "explore",
                "manage_models",
                "see_datagroups",
                "see_lookml",
                "see_lookml_dashboards",
                "see_looks",
                "see_pdts",
                "see_queries",
                "see_schedules",
                "see_sql",
                "see_user_dashboards",
                "see_users",
            }

            USAGE_INGEST_REQUIRED_PERMISSIONS = {
                "see_system_activity",
            }

            LookerDashboardSource._set_test_connection_capability(
                test_report,
                permissions,
                SourceCapability.DESCRIPTIONS,
                BASIC_INGEST_REQUIRED_PERMISSIONS,
            )
            LookerDashboardSource._set_test_connection_capability(
                test_report,
                permissions,
                SourceCapability.OWNERSHIP,
                BASIC_INGEST_REQUIRED_PERMISSIONS,
            )
            LookerDashboardSource._set_test_connection_capability(
                test_report,
                permissions,
                SourceCapability.USAGE_STATS,
                USAGE_INGEST_REQUIRED_PERMISSIONS,
            )
        except Exception as e:
            logger.exception(f"Failed to test connection due to {e}")
            test_report.internal_failure = True
            test_report.internal_failure_reason = f"{e}"

            if test_report.basic_connectivity is None:
                test_report.basic_connectivity = CapabilityReport(
                    capable=False, failure_reason=f"{e}"
                )

        return test_report

    @staticmethod
    def _set_test_connection_capability(
        test_report: TestConnectionReport,
        permissions: Set[str],
        perm: SourceCapability,
        required: Set[str],
    ) -> None:
        assert test_report.capability_report is not None

        if required.issubset(permissions):
            test_report.capability_report[perm] = CapabilityReport(capable=True)
        else:
            missing = required - permissions
            test_report.capability_report[perm] = CapabilityReport(
                capable=False,
                error_message=f"Missing required permissions: {', '.join(missing)}",
            )

    @staticmethod
    def _extract_view_from_field(field: str) -> str:
        assert (
            field.count(".") == 1
        ), f"Error: A field must be prefixed by a view name, field is: {field}"
        return field.split(".")[0]

    def _get_views_from_fields(self, fields: List[str]) -> List[str]:
        field_set = set(fields)
        views = set()
        for field_name in field_set:
            try:
                view_name = self._extract_view_from_field(field_name)
                views.add(view_name)
            except AssertionError:
                self.reporter.report_warning(
                    title="Failed to Extract View Name from Field",
                    message="The field was not prefixed by a view name. This can happen when the field references another dynamic field.",
                    context=f"Field Name: {field_name}",
                )
                continue

        return list(views)

    def _get_input_fields_from_query(
        self, query: Optional[Query]
    ) -> List[InputFieldElement]:
        if query is None:
            return []
        result = []

        # query.dynamic_fields can contain:
        # - looker table calculations: https://docs.looker.com/exploring-data/using-table-calculations
        # - looker custom measures: https://docs.looker.com/de/exploring-data/adding-fields/custom-measure
        # - looker custom dimensions: https://docs.looker.com/exploring-data/adding-fields/custom-measure#creating_a_custom_dimension_using_a_looker_expression
        try:
            dynamic_fields = json.loads(
                query.dynamic_fields if query.dynamic_fields is not None else "[]"
            )
        except JSONDecodeError as e:
            logger.warning(
                f"Json load failed on loading dynamic field with error: {e}. The field value was: {query.dynamic_fields}"
            )
            dynamic_fields = []

        for field in dynamic_fields:
            if "table_calculation" in field:
                result.append(
                    InputFieldElement(
                        name=field["table_calculation"],
                        view_field=ViewField(
                            name=field["table_calculation"],
                            label=field.get("label"),
                            field_type=ViewFieldType.UNKNOWN,
                            type="string",
                            description="",
                        ),
                    )
                )
            if "measure" in field:
                # for measure, we can also make sure to index the underlying field that the measure uses
                based_on = field.get("based_on")
                if based_on is not None:
                    result.append(
                        InputFieldElement(
                            based_on,
                            view_field=None,
                            model=query.model,
                            explore=query.view,
                        )
                    )
                result.append(
                    InputFieldElement(
                        name=field["measure"],
                        view_field=ViewField(
                            name=field["measure"],
                            label=field.get("label"),
                            field_type=ViewFieldType.MEASURE,
                            type="string",
                            description="",
                        ),
                    )
                )
            if "dimension" in field:
                result.append(
                    InputFieldElement(
                        name=field["dimension"],
                        view_field=ViewField(
                            name=field["dimension"],
                            label=field.get("label"),
                            field_type=ViewFieldType.DIMENSION,
                            type="string",
                            description="",
                        ),
                    )
                )

        # A query uses fields defined in explores, find the metadata about that field
        fields: Sequence[str] = query.fields if query.fields is not None else []
        for field in fields:
            if field is None:
                continue

            # we haven't loaded in metadata about the explore yet, so we need to wait until explores are populated
            # later to fetch this
            result.append(
                InputFieldElement(
                    name=field, view_field=None, model=query.model, explore=query.view
                )
            )

        # A query uses fields for filtering, and those fields are defined in views, find the views those fields use
        filters: MutableMapping[str, Any] = (
            query.filters if query.filters is not None else {}
        )
        for field in filters.keys():
            if field is None:
                continue

            # we haven't loaded in metadata about the explore yet, so we need to wait until explores are populated
            # later to fetch this
            result.append(
                InputFieldElement(
                    name=field, view_field=None, model=query.model, explore=query.view
                )
            )

        return result

    def add_reachable_explore(self, model: str, explore: str, via: str) -> None:
        if (model, explore) not in self.reachable_explores:
            self.reachable_explores[(model, explore)] = []

        self.reachable_explores[(model, explore)].append(via)

    def _get_looker_dashboard_element(  # noqa: C901
        self, element: DashboardElement
    ) -> Optional[LookerDashboardElement]:
        # Dashboard elements can use raw usage_queries against explores
        explores: List[str]
        input_fields: List[InputFieldElement]

        if element.id is None:
            raise ValueError("Element ID can't be None")

        if element.query is not None:
            input_fields = self._get_input_fields_from_query(element.query)

            # Get the explore from the view directly
            explores = [element.query.view] if element.query.view is not None else []
            logger.debug(
                f"Element {element.title}: Explores added via query: {explores}"
            )
            for exp in explores:
                self.add_reachable_explore(
                    model=element.query.model,
                    explore=exp,
                    via=f"look:{element.look_id}:query:{element.dashboard_id}",
                )

            return LookerDashboardElement(
                id=element.id,
                title=element.title if element.title is not None else "",
                type=element.type,
                description=element.subtitle_text,
                look_id=None,
                query_slug=element.query.slug if element.query.slug is not None else "",
                upstream_explores=[
                    LookerExplore(model_name=element.query.model, name=exp)
                    for exp in explores
                ],
                input_fields=input_fields,
                owner=None,
            )

        # Dashboard elements can *alternatively* link to an existing look
        elif element.look is not None:
            # we pick from element title by default, falling back to look title.
            title: str = (
                element.title
                if element.title is not None and element.title != ""
                else element.look.title
                if element.look.title is not None
                else ""
            )
            if element.look.query is not None:
                input_fields = self._get_input_fields_from_query(element.look.query)
                if element.look.query.view is not None:
                    explores = [element.look.query.view]
                logger.debug(f"Element {title}: Explores added via look: {explores}")
                for exp in explores:
                    self.add_reachable_explore(
                        model=element.look.query.model,
                        explore=exp,
                        via=f"Look:{element.look_id}:query:{element.dashboard_id}",
                    )

                if element.look.query and element.look.query.slug:
                    slug = element.look.query.slug
                else:
                    slug = ""

                return LookerDashboardElement(
                    id=element.id,
                    title=title,
                    type=element.type,
                    description=element.subtitle_text,
                    look_id=element.look_id,
                    query_slug=slug,
                    upstream_explores=[
                        LookerExplore(model_name=element.look.query.model, name=exp)
                        for exp in explores
                    ],
                    input_fields=input_fields,
                    folder=(
                        self._get_looker_folder(element.look.folder)
                        if element.look.folder
                        else None
                    ),
                    folder_path=(
                        self._get_folder_path(element.look.folder, self.looker_api)
                        if element.look.folder
                        else None
                    ),
                    owner=self._get_looker_user(element.look.user_id),
                )

        # Failing the above two approaches, pick out details from result_maker
        elif element.result_maker is not None:
            model: str = ""
            input_fields = []

            explores = []
            if element.result_maker.query is not None:
                model = element.result_maker.query.model
                if element.result_maker.query.view is not None:
                    explores.append(element.result_maker.query.view)
                input_fields = self._get_input_fields_from_query(
                    element.result_maker.query
                )

                logger.debug(
                    f"Element {element.title}: Explores added via result_maker: {explores}"
                )

                for exp in explores:
                    self.add_reachable_explore(
                        model=element.result_maker.query.model,
                        explore=exp,
                        via=f"Look:{element.look_id}:resultmaker:query",
                    )

            # In addition to the query, filters can point to fields as well
            assert element.result_maker.filterables is not None
            for filterable in element.result_maker.filterables:
                if filterable.view is not None and filterable.model is not None:
                    model = filterable.model
                    explores.append(filterable.view)
                    self.add_reachable_explore(
                        model=filterable.model,
                        explore=filterable.view,
                        via=f"Look:{element.look_id}:resultmaker:filterable",
                    )
                listen = filterable.listen
                query = element.result_maker.query
                if listen is not None:
                    for listener in listen:
                        if listener.field is not None:
                            input_fields.append(
                                InputFieldElement(
                                    listener.field,
                                    view_field=None,
                                    model=query.model if query is not None else "",
                                    explore=query.view if query is not None else "",
                                )
                            )

            explores = sorted(list(set(explores)))  # dedup the list of views

            return LookerDashboardElement(
                id=element.id,
                title=element.title if element.title is not None else "",
                type=element.type,
                description=element.subtitle_text,
                look_id=element.look_id,
                query_slug=(
                    element.result_maker.query.slug
                    if element.result_maker.query is not None
                    and element.result_maker.query.slug is not None
                    else ""
                ),
                upstream_explores=[
                    LookerExplore(model_name=model, name=exp) for exp in explores
                ],
                input_fields=input_fields,
                owner=None,
            )

        logger.debug(f"Element {element.title}: Unable to parse LookerDashboardElement")
        return None

    def _get_chart_type(
        self, dashboard_element: LookerDashboardElement
    ) -> Optional[str]:
        type_mapping = {
            "looker_column": ChartTypeClass.BAR,
            "looker_scatter": ChartTypeClass.SCATTER,
            "looker_line": ChartTypeClass.LINE,
            "looker_area": ChartTypeClass.AREA,
            "looker_pie": ChartTypeClass.PIE,
            "looker_donut_multiples": ChartTypeClass.PIE,
            "looker_funnel": ChartTypeClass.BAR,
            "looker_timeline": ChartTypeClass.BAR,
            "looker_waterfall": ChartTypeClass.BAR,
            "text": ChartTypeClass.TEXT,
            "single_value": ChartTypeClass.TEXT,
            "looker_single_record": ChartTypeClass.TABLE,
            "table": ChartTypeClass.TABLE,
            "looker_grid": ChartTypeClass.TABLE,
            "looker_map": None,
            "looker_geo_coordinates": None,
            "looker_geo_choropleth": None,
            "looker_boxplot": ChartTypeClass.BOX_PLOT,
            "vis": None,
        }
        type_str = dashboard_element.type
        if not type_str:
            self.reporter.info(
                title="Unrecognized Chart Type",
                message="Chart is missing a chart type.",
                context=f"Chart Id: {dashboard_element.id}",
                log=False,
            )
            return None
        try:
            chart_type = type_mapping[type_str]
        except KeyError:
            self.reporter.info(
                title="Unrecognized Chart Type",
                message=f"Chart type {type_str} is not recognized. Setting to None",
                context=f"Chart Id: {dashboard_element.id}",
                log=False,
            )
            chart_type = None

        return chart_type

    def _get_folder_browse_path_v2_entries(
        self, folder: LookerFolder, include_current_folder: bool = True
    ) -> Iterable[BrowsePathEntryClass]:
        for ancestor in self.looker_api.folder_ancestors(folder_id=folder.id):
            assert ancestor.id
            urn = self._gen_folder_key(ancestor.id).as_urn()
            yield BrowsePathEntryClass(id=urn, urn=urn)

        urn = self._gen_folder_key(folder.id).as_urn()
        if include_current_folder:
            yield BrowsePathEntryClass(id=urn, urn=urn)

    def _create_platform_instance_aspect(
        self,
    ) -> DataPlatformInstance:

        assert (
            self.source_config.platform_name
        ), "Platform name is not set in the configuration."
        assert (
            self.source_config.platform_instance
        ), "Platform instance is not set in the configuration."

        return DataPlatformInstance(
            platform=builder.make_data_platform_urn(self.source_config.platform_name),
            instance=builder.make_dataplatform_instance_urn(
                platform=self.source_config.platform_name,
                instance=self.source_config.platform_instance,
            ),
        )

    def _make_chart_urn(self, element_id: str) -> str:
        platform_instance: Optional[str] = None
        if self.source_config.include_platform_instance_in_urns:
            platform_instance = self.source_config.platform_instance

        return builder.make_chart_urn(
            name=element_id,
            platform=self.source_config.platform_name,
            platform_instance=platform_instance,
        )

    def _make_chart_metadata_events(
        self,
        dashboard_element: LookerDashboardElement,
        dashboard: Optional[
            LookerDashboard
        ],  # dashboard will be None if this is a standalone look
    ) -> List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]]:
        chart_urn = self._make_chart_urn(
            element_id=dashboard_element.get_urn_element_id()
        )
        chart_snapshot = ChartSnapshot(
            urn=chart_urn,
            aspects=[Status(removed=False)],
        )
        browse_path_v2: Optional[BrowsePathsV2Class] = None

        chart_type = self._get_chart_type(dashboard_element)
        chart_info = ChartInfoClass(
            type=chart_type,
            description=dashboard_element.description or "",
            title=dashboard_element.title or "",
            lastModified=ChangeAuditStamps(),
            chartUrl=dashboard_element.url(self.source_config.external_base_url or ""),
            inputs=dashboard_element.get_view_urns(self.source_config),
            customProperties={
                "upstream_fields": (
                    ",".join(
                        sorted({field.name for field in dashboard_element.input_fields})
                    )
                    if dashboard_element.input_fields
                    else ""
                )
            },
        )
        chart_snapshot.aspects.append(chart_info)

        if (
            dashboard
            and dashboard.folder_path is not None
            and dashboard.folder is not None
        ):
            browse_path = BrowsePathsClass(
                paths=[f"/Folders/{dashboard.folder_path}/{dashboard.title}"]
            )
            chart_snapshot.aspects.append(browse_path)

            dashboard_urn = self.make_dashboard_urn(dashboard)
            browse_path_v2 = BrowsePathsV2Class(
                path=[
                    BrowsePathEntryClass("Folders"),
                    *self._get_folder_browse_path_v2_entries(dashboard.folder),
                    BrowsePathEntryClass(id=dashboard_urn, urn=dashboard_urn),
                ],
            )
        elif (
            dashboard is None
            and dashboard_element.folder_path is not None
            and dashboard_element.folder is not None
        ):  # independent look
            browse_path = BrowsePathsClass(
                paths=[f"/Folders/{dashboard_element.folder_path}"]
            )
            chart_snapshot.aspects.append(browse_path)
            browse_path_v2 = BrowsePathsV2Class(
                path=[
                    BrowsePathEntryClass("Folders"),
                    *self._get_folder_browse_path_v2_entries(dashboard_element.folder),
                ],
            )

        if dashboard is not None:
            ownership = self.get_ownership(dashboard)
            if ownership is not None:
                chart_snapshot.aspects.append(ownership)
        elif dashboard is None and dashboard_element is not None:
            ownership = self.get_ownership(dashboard_element)
            if ownership is not None:
                chart_snapshot.aspects.append(ownership)

        chart_mce = MetadataChangeEvent(proposedSnapshot=chart_snapshot)

        proposals: List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]] = [
            chart_mce,
            MetadataChangeProposalWrapper(
                entityUrn=chart_urn,
                aspect=SubTypesClass(typeNames=[BIAssetSubTypes.LOOKER_LOOK]),
            ),
        ]

        if self.source_config.include_platform_instance_in_urns:
            proposals.append(
                MetadataChangeProposalWrapper(
                    entityUrn=chart_urn,
                    aspect=self._create_platform_instance_aspect(),
                ),
            )

        # If extracting embeds is enabled, produce an MCP for embed URL.
        if (
            self.source_config.extract_embed_urls
            and self.source_config.external_base_url
        ):
            maybe_embed_url = dashboard_element.embed_url(
                self.source_config.external_base_url
            )
            if maybe_embed_url:
                proposals.append(
                    create_embed_mcp(
                        chart_snapshot.urn,
                        maybe_embed_url,
                    )
                )

        if dashboard is None and dashboard_element.folder:
            container = ContainerClass(
                container=self._gen_folder_key(dashboard_element.folder.id).as_urn(),
            )
            proposals.append(
                MetadataChangeProposalWrapper(entityUrn=chart_urn, aspect=container)
            )

        if browse_path_v2:
            proposals.append(
                MetadataChangeProposalWrapper(
                    entityUrn=chart_urn, aspect=browse_path_v2
                )
            )

        return proposals

    def _make_dashboard_metadata_events(
        self, looker_dashboard: LookerDashboard, chart_urns: List[str]
    ) -> List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]]:
        dashboard_urn = self.make_dashboard_urn(looker_dashboard)
        dashboard_snapshot = DashboardSnapshot(
            urn=dashboard_urn,
            aspects=[],
        )
        browse_path_v2: Optional[BrowsePathsV2Class] = None
        dashboard_info = DashboardInfoClass(
            description=looker_dashboard.description or "",
            title=looker_dashboard.title,
            charts=chart_urns,
            lastModified=self._get_change_audit_stamps(looker_dashboard),
            dashboardUrl=looker_dashboard.url(self.source_config.external_base_url),
        )

        dashboard_snapshot.aspects.append(dashboard_info)
        if (
            looker_dashboard.folder_path is not None
            and looker_dashboard.folder is not None
        ):
            browse_path = BrowsePathsClass(
                paths=[f"/Folders/{looker_dashboard.folder_path}"]
            )
            browse_path_v2 = BrowsePathsV2Class(
                path=[
                    BrowsePathEntryClass("Folders"),
                    *self._get_folder_browse_path_v2_entries(looker_dashboard.folder),
                ],
            )
            dashboard_snapshot.aspects.append(browse_path)

        ownership = self.get_ownership(looker_dashboard)
        if ownership is not None:
            dashboard_snapshot.aspects.append(ownership)

        dashboard_snapshot.aspects.append(Status(removed=looker_dashboard.is_deleted))

        dashboard_mce = MetadataChangeEvent(proposedSnapshot=dashboard_snapshot)

        proposals: List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]] = [
            dashboard_mce
        ]

        if looker_dashboard.folder is not None:
            container = ContainerClass(
                container=self._gen_folder_key(looker_dashboard.folder.id).as_urn(),
            )
            proposals.append(
                MetadataChangeProposalWrapper(entityUrn=dashboard_urn, aspect=container)
            )

        if browse_path_v2:
            proposals.append(
                MetadataChangeProposalWrapper(
                    entityUrn=dashboard_urn, aspect=browse_path_v2
                )
            )

        # If extracting embeds is enabled, produce an MCP for embed URL.
        if (
            self.source_config.extract_embed_urls
            and self.source_config.external_base_url
        ):
            proposals.append(
                create_embed_mcp(
                    dashboard_snapshot.urn,
                    looker_dashboard.embed_url(self.source_config.external_base_url),
                )
            )

        if self.source_config.include_platform_instance_in_urns:
            proposals.append(
                MetadataChangeProposalWrapper(
                    entityUrn=dashboard_urn,
                    aspect=self._create_platform_instance_aspect(),
                )
            )

        return proposals

    def _make_dashboard_urn(self, looker_dashboard_name_part: str) -> str:
        # Note that `looker_dashboard_name_part` will like be `dashboard.1234`.
        platform_instance: Optional[str] = None
        if self.source_config.include_platform_instance_in_urns:
            platform_instance = self.source_config.platform_instance

        return builder.make_dashboard_urn(
            name=looker_dashboard_name_part,
            platform=self.source_config.platform_name,
            platform_instance=platform_instance,
        )

    def make_dashboard_urn(self, looker_dashboard: LookerDashboard) -> str:
        return self._make_dashboard_urn(looker_dashboard.get_urn_dashboard_id())

    def _make_explore_metadata_events(
        self,
    ) -> Iterable[
        Union[MetadataChangeEvent, MetadataChangeProposalWrapper, MetadataWorkUnit]
    ]:
        if not self.source_config.emit_used_explores_only:
            explores_to_fetch = list(self.list_all_explores())
        else:
            # We don't keep track of project names for each explore right now.
            # Because project names are just used for a custom property, it's
            # fine to set them to None.
            # TODO: Track project names for each explore.
            explores_to_fetch = [
                (None, model, explore)
                for (model, explore) in self.reachable_explores.keys()
            ]
        explores_to_fetch.sort()

        processed_models: List[str] = []

        for project_name, model, _ in explores_to_fetch:
            if model not in processed_models:
                model_key = gen_model_key(self.source_config, model)
                yield from gen_containers(
                    container_key=model_key,
                    name=model,
                    sub_types=[BIContainerSubTypes.LOOKML_MODEL],
                    extra_properties=(
                        {"project": project_name} if project_name is not None else None
                    ),
                )
                yield MetadataChangeProposalWrapper(
                    entityUrn=model_key.as_urn(),
                    aspect=BrowsePathsV2Class(
                        path=[BrowsePathEntryClass("Explore")],
                    ),
                )

                processed_models.append(model)

        self.reporter.total_explores = len(explores_to_fetch)
        for future in BackpressureAwareExecutor.map(
            self.fetch_one_explore,
            ((model, explore) for (_project, model, explore) in explores_to_fetch),
            max_workers=self.source_config.max_threads,
        ):
            events, explore_id, start_time, end_time = future.result()
            self.reporter.explores_scanned += 1
            yield from events
            self.reporter.report_upstream_latency(start_time, end_time)
            logger.debug(
                f"Running time of fetch_one_explore for {explore_id}: {(end_time - start_time).total_seconds()}"
            )

    def list_all_explores(self) -> Iterable[Tuple[Optional[str], str, str]]:
        # returns a list of (model, explore) tuples

        for model in self.looker_api.all_lookml_models():
            if model.name is None or model.explores is None:
                continue
            for explore in model.explores:
                if explore.name is None:
                    continue
                yield (model.project_name, model.name, explore.name)

    def fetch_one_explore(
        self, model: str, explore: str
    ) -> Tuple[
        List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]],
        str,
        datetime.datetime,
        datetime.datetime,
    ]:
        start_time = datetime.datetime.now()
        events: List[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]] = []
        looker_explore = self.explore_registry.get_explore(model, explore)
        if looker_explore is not None:
            events = (
                looker_explore._to_metadata_events(
                    self.source_config,
                    self.reporter,
                    self.source_config.external_base_url or self.source_config.base_url,
                    self.source_config.extract_embed_urls,
                )
                or events
            )

        return events, f"{model}:{explore}", start_time, datetime.datetime.now()

    def _extract_event_urn(
        self, event: Union[MetadataChangeEvent, MetadataChangeProposalWrapper]
    ) -> Optional[str]:
        if isinstance(event, MetadataChangeEvent):
            return event.proposedSnapshot.urn
        else:
            return event.entityUrn

    def _emit_folder_as_container(
        self, folder: LookerFolder
    ) -> Iterable[MetadataWorkUnit]:
        if folder.id not in self.processed_folders:
            yield from gen_containers(
                container_key=self._gen_folder_key(folder.id),
                name=folder.name,
                sub_types=[BIContainerSubTypes.LOOKER_FOLDER],
                parent_container_key=(
                    self._gen_folder_key(folder.parent_id) if folder.parent_id else None
                ),
            )
            if folder.parent_id is None:
                yield MetadataChangeProposalWrapper(
                    entityUrn=self._gen_folder_key(folder.id).as_urn(),
                    aspect=BrowsePathsV2Class(
                        path=[BrowsePathEntryClass("Folders")],
                    ),
                ).as_workunit()
            else:
                yield MetadataChangeProposalWrapper(
                    entityUrn=self._gen_folder_key(folder.id).as_urn(),
                    aspect=BrowsePathsV2Class(
                        path=[
                            BrowsePathEntryClass("Folders"),
                            *self._get_folder_browse_path_v2_entries(
                                folder, include_current_folder=False
                            ),
                        ],
                    ),
                ).as_workunit()
            self.processed_folders.append(folder.id)

    def _gen_folder_key(self, folder_id: str) -> LookerFolderKey:
        return LookerFolderKey(
            folder_id=folder_id,
            env=self.source_config.env,
            platform=self.source_config.platform_name,
            instance=self.source_config.platform_instance,
        )

    def _make_dashboard_and_chart_mces(
        self, looker_dashboard: LookerDashboard
    ) -> Iterable[Union[MetadataChangeEvent, MetadataChangeProposalWrapper]]:

        # Step 1: Emit metadata for each Chart inside the Dashboard.
        chart_events = []
        for element in looker_dashboard.dashboard_elements:
            if element.type == "vis":
                chart_events.extend(
                    self._make_chart_metadata_events(element, looker_dashboard)
                )

        yield from chart_events

        # Step 2: Emit metadata events for the Dashboard itself.
        chart_urns: Set[
            str
        ] = set()  # Collect the unique child chart urns for dashboard input lineage.
        for chart_event in chart_events:
            chart_event_urn = self._extract_event_urn(chart_event)
            if chart_event_urn:
                chart_urns.add(chart_event_urn)

        dashboard_events = self._make_dashboard_metadata_events(
            looker_dashboard, list(chart_urns)
        )
        yield from dashboard_events

    def get_ownership(
        self, looker_dashboard_look: Union[LookerDashboard, LookerDashboardElement]
    ) -> Optional[OwnershipClass]:
        if looker_dashboard_look.owner is not None:
            owner_urn = looker_dashboard_look.owner.get_urn(
                self.source_config.strip_user_ids_from_email
            )
            if owner_urn is not None:
                ownership: OwnershipClass = OwnershipClass(
                    owners=[
                        OwnerClass(
                            owner=owner_urn,
                            type=OwnershipTypeClass.DATAOWNER,
                        )
                    ]
                )
                return ownership
        return None

    def _get_change_audit_stamps(
        self, looker_dashboard: LookerDashboard
    ) -> ChangeAuditStamps:
        change_audit_stamp: ChangeAuditStamps = ChangeAuditStamps()
        if looker_dashboard.created_at is not None:
            change_audit_stamp.created.time = round(
                looker_dashboard.created_at.timestamp() * 1000
            )
        if looker_dashboard.owner is not None:
            owner_urn = looker_dashboard.owner.get_urn(
                self.source_config.strip_user_ids_from_email
            )
            if owner_urn:
                change_audit_stamp.created.actor = owner_urn
        if looker_dashboard.last_updated_at is not None:
            change_audit_stamp.lastModified.time = round(
                looker_dashboard.last_updated_at.timestamp() * 1000
            )
        if looker_dashboard.last_updated_by is not None:
            updated_by_urn = looker_dashboard.last_updated_by.get_urn(
                self.source_config.strip_user_ids_from_email
            )
            if updated_by_urn:
                change_audit_stamp.lastModified.actor = updated_by_urn
        if (
            looker_dashboard.is_deleted
            and looker_dashboard.deleted_by is not None
            and looker_dashboard.deleted_at is not None
        ):
            deleter_urn = looker_dashboard.deleted_by.get_urn(
                self.source_config.strip_user_ids_from_email
            )
            if deleter_urn:
                change_audit_stamp.deleted = AuditStamp(
                    actor=deleter_urn,
                    time=round(looker_dashboard.deleted_at.timestamp() * 1000),
                )

        return change_audit_stamp

    def _get_looker_folder(self, folder: Union[Folder, FolderBase]) -> LookerFolder:
        assert folder.id
        return LookerFolder(id=folder.id, name=folder.name, parent_id=folder.parent_id)

    def _get_folder_path(self, folder: FolderBase, client: LookerAPI) -> str:
        assert folder.id
        ancestors = [
            ancestor.name for ancestor in client.folder_ancestors(folder_id=folder.id)
        ]
        return "/".join(ancestors + [folder.name])

    def _get_looker_dashboard(self, dashboard: Dashboard) -> LookerDashboard:
        self.reporter.accessed_dashboards += 1
        if dashboard.folder is None:
            logger.debug(f"{dashboard.id} has no folder")
        dashboard_folder = None
        if dashboard.folder is not None:
            dashboard_folder = self._get_looker_folder(dashboard.folder)
        dashboard_elements: List[LookerDashboardElement] = []
        elements = (
            dashboard.dashboard_elements
            if dashboard.dashboard_elements is not None
            else []
        )
        for element in elements:
            self.reporter.report_charts_scanned()
            if element.id is not None and not self.source_config.chart_pattern.allowed(
                element.id
            ):
                self.reporter.report_charts_dropped(element.id)
                continue
            if element.look_id is not None:
                # Keeping track of reachable element from Dashboard
                # Later we need to ingest looks which are not reachable from any dashboard
                self.reachable_look_registry.add(element.look_id)
            looker_dashboard_element = self._get_looker_dashboard_element(element)
            if looker_dashboard_element is not None:
                dashboard_elements.append(looker_dashboard_element)

        if dashboard.id is None or dashboard.title is None:
            raise ValueError("Both dashboard ID and title are None")

        looker_dashboard = LookerDashboard(
            id=dashboard.id,
            title=dashboard.title,
            description=dashboard.description,
            dashboard_elements=dashboard_elements,
            created_at=dashboard.created_at,
            is_deleted=dashboard.deleted if dashboard.deleted is not None else False,
            is_hidden=dashboard.hidden if dashboard.hidden is not None else False,
            folder=dashboard_folder,
            folder_path=(
                self._get_folder_path(dashboard.folder, self.looker_api)
                if dashboard.folder
                else None
            ),
            owner=self._get_looker_user(dashboard.user_id),
            strip_user_ids_from_email=self.source_config.strip_user_ids_from_email,
            last_updated_at=dashboard.updated_at,
            last_updated_by=self._get_looker_user(dashboard.last_updater_id),
            deleted_at=dashboard.deleted_at,
            deleted_by=self._get_looker_user(dashboard.deleter_id),
            favorite_count=dashboard.favorite_count,
            view_count=dashboard.view_count,
            last_viewed_at=dashboard.last_viewed_at,
        )
        return looker_dashboard

    def _get_looker_user(self, user_id: Optional[str]) -> Optional[LookerUser]:
        user = (
            self.user_registry.get_by_id(
                user_id,
            )
            if self.source_config.extract_owners and user_id is not None
            else None
        )

        if user is not None and self.source_config.extract_owners:
            # Keep track of how many user ids we were able to resolve
            self.reporter.resolved_user_ids += 1
            if user.email is None:
                self.reporter.email_ids_missing += 1

        return user

    def process_metrics_dimensions_and_fields_for_dashboard(
        self, dashboard: LookerDashboard
    ) -> List[MetadataWorkUnit]:
        chart_mcps = [
            self._make_metrics_dimensions_chart_mcp(element)
            for element in dashboard.dashboard_elements
        ]
        dashboard_mcp = self._make_metrics_dimensions_dashboard_mcp(dashboard)

        mcps = chart_mcps
        mcps.append(dashboard_mcp)

        workunits = [mcp.as_workunit() for mcp in mcps]

        return workunits

    def _input_fields_from_dashboard_element(
        self, dashboard_element: LookerDashboardElement
    ) -> List[InputFieldClass]:
        input_fields = (
            dashboard_element.input_fields
            if dashboard_element.input_fields is not None
            else []
        )

        fields_for_mcp = []

        # enrich the input_fields with the fully hydrated ViewField from the now fetched explores
        for input_field in input_fields:
            entity_urn = self._make_chart_urn(
                element_id=dashboard_element.get_urn_element_id()
            )
            view_field_for_reference = input_field.view_field

            if input_field.view_field is None:
                explore = self.explore_registry.get_explore(
                    input_field.model, input_field.explore
                )
                if explore is not None:
                    # add this to the list of explores to finally generate metadata for
                    self.add_reachable_explore(
                        input_field.model, input_field.explore, entity_urn
                    )
                    entity_urn = explore.get_explore_urn(self.source_config)
                    explore_fields = (
                        explore.fields if explore.fields is not None else []
                    )

                    relevant_field = next(
                        (
                            field
                            for field in explore_fields
                            if field.name == input_field.name
                        ),
                        None,
                    )
                    if relevant_field is not None:
                        view_field_for_reference = relevant_field

            if view_field_for_reference and view_field_for_reference.name:
                fields_for_mcp.append(
                    InputFieldClass(
                        schemaFieldUrn=builder.make_schema_field_urn(
                            entity_urn, view_field_for_reference.name
                        ),
                        schemaField=LookerUtil.view_field_to_schema_field(
                            view_field_for_reference,
                            self.reporter,
                            self.source_config.tag_measures_and_dimensions,
                        ),
                    )
                )

        return fields_for_mcp

    def _make_metrics_dimensions_dashboard_mcp(
        self, dashboard: LookerDashboard
    ) -> MetadataChangeProposalWrapper:
        dashboard_urn = self.make_dashboard_urn(dashboard)
        all_fields = []
        for dashboard_element in dashboard.dashboard_elements:
            all_fields.extend(
                self._input_fields_from_dashboard_element(dashboard_element)
            )

        input_fields_aspect = InputFieldsClass(fields=all_fields)

        return MetadataChangeProposalWrapper(
            entityUrn=dashboard_urn,
            aspect=input_fields_aspect,
        )

    def _make_metrics_dimensions_chart_mcp(
        self, dashboard_element: LookerDashboardElement
    ) -> MetadataChangeProposalWrapper:
        chart_urn = self._make_chart_urn(
            element_id=dashboard_element.get_urn_element_id()
        )
        input_fields_aspect = InputFieldsClass(
            fields=self._input_fields_from_dashboard_element(dashboard_element)
        )

        return MetadataChangeProposalWrapper(
            entityUrn=chart_urn,
            aspect=input_fields_aspect,
        )

    def process_dashboard(
        self, dashboard_id: str, fields: List[str]
    ) -> Tuple[
        List[MetadataWorkUnit],
        Optional[looker_usage.LookerDashboardForUsage],
        str,
        datetime.datetime,
        datetime.datetime,
    ]:
        start_time = datetime.datetime.now()
        assert dashboard_id is not None
        if not self.source_config.dashboard_pattern.allowed(dashboard_id):
            self.reporter.report_dashboards_dropped(dashboard_id)
            return [], None, dashboard_id, start_time, datetime.datetime.now()
        try:
            dashboard_object: Dashboard = self.looker_api.dashboard(
                dashboard_id=dashboard_id,
                fields=fields,
            )
        except (SDKError, DeserializeError) as e:
            # A looker dashboard could be deleted in between the list and the get
            self.reporter.report_warning(
                title="Failed to fetch dashboard from the Looker API",
                message="Error occurred while attempting to loading dashboard from Looker API. Skipping.",
                context=f"Dashboard ID: {dashboard_id}",
                exc=e,
            )
            return [], None, dashboard_id, start_time, datetime.datetime.now()

        if self.source_config.skip_personal_folders:
            if dashboard_object.folder is not None and (
                dashboard_object.folder.is_personal
                or dashboard_object.folder.is_personal_descendant
            ):
                self.reporter.info(
                    title="Dropped Dashboard",
                    message="Dropped due to being a personal folder",
                    context=f"Dashboard ID: {dashboard_id}",
                )
                self.reporter.report_dashboards_dropped(dashboard_id)
                return [], None, dashboard_id, start_time, datetime.datetime.now()

        looker_dashboard = self._get_looker_dashboard(dashboard_object)

        workunits = []
        if (
            looker_dashboard.folder_path is not None
            and not self.source_config.folder_path_pattern.allowed(
                looker_dashboard.folder_path
            )
        ):
            logger.debug(
                f"Folder path {looker_dashboard.folder_path} is denied in folder_path_pattern"
            )
            return [], None, dashboard_id, start_time, datetime.datetime.now()

        if looker_dashboard.folder:
            workunits += list(
                self._get_folder_and_ancestors_workunits(looker_dashboard.folder)
            )

        mces = self._make_dashboard_and_chart_mces(looker_dashboard)
        workunits += [
            (
                MetadataWorkUnit(id=f"looker-{mce.proposedSnapshot.urn}", mce=mce)
                if isinstance(mce, MetadataChangeEvent)
                else MetadataWorkUnit(
                    id=f"looker-{mce.aspectName}-{mce.entityUrn}", mcp=mce
                )
            )
            for mce in mces
        ]

        # add on metrics, dimensions, fields events
        metric_dim_workunits = self.process_metrics_dimensions_and_fields_for_dashboard(
            looker_dashboard
        )

        workunits.extend(metric_dim_workunits)

        self.reporter.report_dashboards_scanned()

        # generate usage tracking object
        dashboard_usage = looker_usage.LookerDashboardForUsage.from_dashboard(
            dashboard_object
        )

        return (
            workunits,
            dashboard_usage,
            dashboard_id,
            start_time,
            datetime.datetime.now(),
        )

    def _get_folder_and_ancestors_workunits(
        self, folder: LookerFolder
    ) -> Iterable[MetadataWorkUnit]:
        for ancestor_folder in self.looker_api.folder_ancestors(folder.id):
            yield from self._emit_folder_as_container(
                self._get_looker_folder(ancestor_folder)
            )
        yield from self._emit_folder_as_container(folder)

    def extract_usage_stat(
        self, looker_dashboards: List[looker_usage.LookerDashboardForUsage]
    ) -> List[MetadataChangeProposalWrapper]:
        looks: List[looker_usage.LookerChartForUsage] = []
        # filter out look from all dashboard
        for dashboard in looker_dashboards:
            if dashboard.looks is None:
                continue
            looks.extend(dashboard.looks)

        # dedup looks
        looks = list({str(look.id): look for look in looks}.values())

        # Keep stat generators to generate entity stat aspect later
        stat_generator_config: looker_usage.StatGeneratorConfig = (
            looker_usage.StatGeneratorConfig(
                looker_api_wrapper=self.looker_api,
                looker_user_registry=self.user_registry,
                interval=self.source_config.extract_usage_history_for_interval,
                strip_user_ids_from_email=self.source_config.strip_user_ids_from_email,
                max_threads=self.source_config.max_threads,
            )
        )

        dashboard_usage_generator = looker_usage.create_dashboard_stat_generator(
            stat_generator_config,
            self.reporter,
            self._make_dashboard_urn,
            looker_dashboards,
        )

        chart_usage_generator = looker_usage.create_chart_stat_generator(
            stat_generator_config,
            self.reporter,
            self._make_chart_urn,
            looks,
        )

        mcps: List[MetadataChangeProposalWrapper] = []
        for usage_stat_generator in [dashboard_usage_generator, chart_usage_generator]:
            for mcp in usage_stat_generator.generate_usage_stat_mcps():
                mcps.append(mcp)

        return mcps

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.source_config, self.ctx
            ).workunit_processor,
        ]

    def emit_independent_looks_mcp(
        self, dashboard_element: LookerDashboardElement
    ) -> Iterable[MetadataWorkUnit]:
        if dashboard_element.folder:  # independent look
            yield from self._get_folder_and_ancestors_workunits(
                dashboard_element.folder
            )

        yield from auto_workunit(
            stream=self._make_chart_metadata_events(
                dashboard_element=dashboard_element,
                dashboard=None,
            )
        )

        yield from auto_workunit(
            [
                self._make_metrics_dimensions_chart_mcp(
                    dashboard_element,
                )
            ]
        )

    def extract_independent_looks(self) -> Iterable[MetadataWorkUnit]:
        """
        Emit MetadataWorkUnit for looks which are not part of any Dashboard
        """
        if self.source_config.extract_independent_looks is False:
            return

        self.reporter.report_stage_start("extract_independent_looks")

        logger.debug("Extracting looks not part of Dashboard")
        look_fields: List[str] = [
            "id",
            "title",
            "description",
            "query_id",
            "folder",
            "user_id",
        ]
        query_fields: List[str] = [
            "id",
            "view",
            "model",
            "dynamic_fields",
            "filters",
            "fields",
            "slug",
        ]

        all_looks: List[Look] = self.looker_api.all_looks(
            fields=look_fields, soft_deleted=self.source_config.include_deleted
        )
        for look in all_looks:
            if look.id in self.reachable_look_registry:
                # This look is reachable from the Dashboard
                continue

            if look.query_id is None:
                logger.info(f"query_id is None for look {look.title}({look.id})")
                continue

            if self.source_config.skip_personal_folders:
                if look.folder is not None and (
                    look.folder.is_personal or look.folder.is_personal_descendant
                ):
                    self.reporter.info(
                        title="Dropped Look",
                        message="Dropped due to being a personal folder",
                        context=f"Look ID: {look.id}",
                    )

                    assert look.id, "Looker id is null"
                    self.reporter.report_charts_dropped(look.id)
                    continue

            if look.id is not None:
                query: Optional[Query] = self.looker_api.get_look(
                    look.id, fields=["query"]
                ).query
                # Only include fields that are in the query_fields list
                query = Query(
                    **{
                        key: getattr(query, key)
                        for key in query_fields
                        if hasattr(query, key)
                    }
                )

            dashboard_element: Optional[
                LookerDashboardElement
            ] = self._get_looker_dashboard_element(
                DashboardElement(
                    id=f"looks_{look.id}",  # to avoid conflict with non-standalone looks (element.id prefixes),
                    # we add the "looks_" prefix to look.id.
                    title=look.title,
                    subtitle_text=look.description,
                    look_id=look.id,
                    dashboard_id=None,  # As this is an independent look
                    look=LookWithQuery(
                        query=query, folder=look.folder, user_id=look.user_id
                    ),
                ),
            )

            if dashboard_element is not None:
                logger.debug(f"Emitting MCPS for look {look.title}({look.id})")
                yield from self.emit_independent_looks_mcp(
                    dashboard_element=dashboard_element
                )

        self.reporter.report_stage_end("extract_independent_looks")

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        self.reporter.report_stage_start("list_dashboards")
        dashboards = self.looker_api.all_dashboards(fields="id")
        deleted_dashboards = (
            self.looker_api.search_dashboards(fields="id", deleted="true")
            if self.source_config.include_deleted
            else []
        )
        if deleted_dashboards != []:
            logger.debug(f"Deleted Dashboards = {deleted_dashboards}")

        dashboard_ids = [dashboard_base.id for dashboard_base in dashboards]
        dashboard_ids.extend(
            [deleted_dashboard.id for deleted_dashboard in deleted_dashboards]
        )
        selected_dashboard_ids: List[Optional[str]] = []
        for id in dashboard_ids:
            if id is None:
                continue
            if not self.source_config.dashboard_pattern.allowed(id):
                self.reporter.report_dashboards_dropped(id)
            else:
                selected_dashboard_ids.append(id)
        dashboard_ids = selected_dashboard_ids
        self.reporter.report_stage_end("list_dashboards")
        self.reporter.report_total_dashboards(len(dashboard_ids))

        # List dashboard fields to extract for processing
        fields = [
            "id",
            "title",
            "dashboard_elements",
            "dashboard_filters",
            "deleted",
            "hidden",
            "description",
            "folder",
            "user_id",
            "created_at",
            "updated_at",
            "last_updater_id",
            "deleted_at",
            "deleter_id",
        ]
        if self.source_config.extract_usage_history:
            fields += [
                "favorite_count",
                "view_count",
                "last_viewed_at",
            ]

        looker_dashboards_for_usage: List[looker_usage.LookerDashboardForUsage] = []

        with self.reporter.report_stage("dashboard_chart_metadata"):
            for job in BackpressureAwareExecutor.map(
                self.process_dashboard,
                (
                    (dashboard_id, fields)
                    for dashboard_id in dashboard_ids
                    if dashboard_id is not None
                ),
                max_workers=self.source_config.max_threads,
            ):
                (
                    work_units,
                    dashboard_usage,
                    dashboard_id,
                    start_time,
                    end_time,
                ) = job.result()
                logger.debug(
                    f"Running time of process_dashboard for {dashboard_id} = {(end_time - start_time).total_seconds()}"
                )
                self.reporter.report_upstream_latency(start_time, end_time)

                yield from work_units
                if dashboard_usage is not None:
                    looker_dashboards_for_usage.append(dashboard_usage)

        if (
            self.source_config.extract_owners
            and self.reporter.resolved_user_ids > 0
            and self.reporter.email_ids_missing == self.reporter.resolved_user_ids
        ):
            # Looks like we tried to extract owners and could not find their email addresses. This is likely a permissions issue
            self.reporter.report_warning(
                "Failed to extract owner emails",
                "Failed to extract owners emails for any dashboards. Please enable the see_users permission for your Looker API key",
            )

        # Extract independent look here, so that explore of this look would get consider in _make_explore_metadata_events
        yield from self.extract_independent_looks()

        self.reporter.report_stage_start("explore_metadata")

        for event in self._make_explore_metadata_events():
            if isinstance(event, MetadataChangeEvent):
                yield MetadataWorkUnit(
                    id=f"looker-{event.proposedSnapshot.urn}", mce=event
                )
            elif isinstance(event, MetadataChangeProposalWrapper):
                yield event.as_workunit()
            elif isinstance(event, MetadataWorkUnit):
                yield event
            else:
                raise Exception(f"Unexpected type of event {event}")
        self.reporter.report_stage_end("explore_metadata")

        if (
            self.source_config.tag_measures_and_dimensions
            and self.reporter.explores_scanned > 0
        ):
            # Emit tag MCEs for measures and dimensions if we produced any explores:
            for tag_mce in LookerUtil.get_tag_mces():
                yield MetadataWorkUnit(
                    id=f"tag-{tag_mce.proposedSnapshot.urn}",
                    mce=tag_mce,
                )

        # Extract usage history is enabled
        if self.source_config.extract_usage_history:
            self.reporter.report_stage_start("usage_extraction")
            usage_mcps: List[MetadataChangeProposalWrapper] = self.extract_usage_stat(
                looker_dashboards_for_usage
            )
            for usage_mcp in usage_mcps:
                yield usage_mcp.as_workunit()
            self.reporter.report_stage_end("usage_extraction")

    def get_report(self) -> SourceReport:
        return self.reporter



from typing import List, Optional, Set, cast

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import (
    TransformerSemantics,
    TransformerSemanticsConfigModel,
)
from datahub.emitter.mce_builder import Aspect, make_term_urn
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.graph.client import DataHubGraph
from datahub.ingestion.transformer.dataset_transformer import TagsToTermTransformer
from datahub.metadata.schema_classes import (
    AuditStampClass,
    GlobalTagsClass,
    GlossaryTermAssociationClass,
    GlossaryTermsClass,
    SchemaMetadataClass,
)


class TagsToTermMapperConfig(TransformerSemanticsConfigModel):
    tags: List[str]


class TagsToTermMapper(TagsToTermTransformer):
    """This transformer maps specified tags to corresponding glossary terms for a dataset."""

    def __init__(self, config: TagsToTermMapperConfig, ctx: PipelineContext):
        super().__init__()
        self.ctx: PipelineContext = ctx
        self.config: TagsToTermMapperConfig = config

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "TagsToTermMapper":
        config = TagsToTermMapperConfig.parse_obj(config_dict)
        return cls(config, ctx)

    @staticmethod
    def _merge_with_server_glossary_terms(
        graph: DataHubGraph,
        urn: str,
        glossary_terms_aspect: Optional[GlossaryTermsClass],
    ) -> Optional[GlossaryTermsClass]:
        if not glossary_terms_aspect or not glossary_terms_aspect.terms:
            # nothing to add, no need to consult server
            return None

        # Merge the transformed terms with existing server terms.
        # The transformed terms takes precedence, which may change the term context.
        server_glossary_terms_aspect = graph.get_glossary_terms(entity_urn=urn)
        if server_glossary_terms_aspect is not None:
            glossary_terms_aspect.terms = list(
                {
                    **{term.urn: term for term in server_glossary_terms_aspect.terms},
                    **{term.urn: term for term in glossary_terms_aspect.terms},
                }.values()
            )

        return glossary_terms_aspect

    @staticmethod
    def get_tags_from_global_tags(global_tags: Optional[GlobalTagsClass]) -> Set[str]:
        """Extracts tags urn from GlobalTagsClass."""
        if not global_tags or not global_tags.tags:
            return set()

        return {tag_assoc.tag for tag_assoc in global_tags.tags}

    @staticmethod
    def get_tags_from_schema_metadata(
        schema_metadata: Optional[SchemaMetadataClass],
    ) -> Set[str]:
        """Extracts globalTags from all fields in SchemaMetadataClass."""
        if not schema_metadata or not schema_metadata.fields:
            return set()
        tags = set()
        for field in schema_metadata.fields:
            if field.globalTags:
                tags.update(
                    TagsToTermMapper.get_tags_from_global_tags(field.globalTags)
                )
        return tags

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[Aspect]
    ) -> Optional[Aspect]:

        in_glossary_terms: Optional[GlossaryTermsClass] = cast(
            Optional[GlossaryTermsClass], aspect
        )

        assert self.ctx.graph
        in_global_tags_aspect: Optional[GlobalTagsClass] = self.ctx.graph.get_tags(
            entity_urn
        )
        in_schema_metadata_aspect: Optional[
            SchemaMetadataClass
        ] = self.ctx.graph.get_schema_metadata(entity_urn)

        if in_global_tags_aspect is None and in_schema_metadata_aspect is None:
            return cast(Aspect, in_glossary_terms)

        global_tags = TagsToTermMapper.get_tags_from_global_tags(in_global_tags_aspect)
        schema_metadata_tags = TagsToTermMapper.get_tags_from_schema_metadata(
            in_schema_metadata_aspect
        )

        # Combine tags from both global and schema level
        combined_tags = global_tags.union(schema_metadata_tags)

        tag_set = set(self.config.tags)
        terms_to_add = set()
        tags_to_delete = set()

        # Check each global tag against the configured tag list and prepare terms
        for full_tag in combined_tags:
            tag_name = full_tag.split("urn:li:tag:")[-1].split(".")[-1].split(":")[0]
            if tag_name in tag_set:
                term_urn = make_term_urn(tag_name)
                terms_to_add.add(term_urn)
                tags_to_delete.add(full_tag)  # Full URN for deletion

        if not terms_to_add:
            return cast(Aspect, in_glossary_terms)  # No new terms to add

        for tag_urn in tags_to_delete:
            self.ctx.graph.remove_tag(tag_urn=tag_urn, resource_urn=entity_urn)

        # Initialize the Glossary Terms properly
        out_glossary_terms = GlossaryTermsClass(
            terms=[GlossaryTermAssociationClass(urn=term) for term in terms_to_add],
            auditStamp=AuditStampClass(
                time=builder.get_sys_time(), actor="urn:li:corpUser:restEmitter"
            ),
        )

        if self.config.semantics == TransformerSemantics.PATCH:
            patch_glossary_terms: Optional[
                GlossaryTermsClass
            ] = TagsToTermMapper._merge_with_server_glossary_terms(
                self.ctx.graph, entity_urn, out_glossary_terms
            )
            return cast(Optional[Aspect], patch_glossary_terms)
        else:
            return cast(Aspect, out_glossary_terms)



import logging
from dataclasses import dataclass, field
from typing import Dict, Optional

import pydantic

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
)

logger = logging.getLogger(__name__)


class Constant:
    """
    keys used in sigma plugin
    """

    # Rest API response key constants
    REFRESH_TOKEN = "refresh_token"
    ACCESS_TOKEN = "access_token"
    ENTRIES = "entries"
    MEMBERID = "memberId"
    FIRSTNAME = "firstName"
    LASTNAME = "lastName"
    EDGES = "edges"
    DEPENDENCIES = "dependencies"
    SOURCE = "source"
    WORKSPACEID = "workspaceId"
    PATH = "path"
    NAME = "name"
    URL = "url"
    ELEMENTID = "elementId"
    ID = "id"
    PARENTID = "parentId"
    TYPE = "type"
    DATASET = "dataset"
    WORKBOOK = "workbook"
    BADGE = "badge"
    NEXTPAGE = "nextPage"

    # Source Config constants
    DEFAULT_API_URL = "https://aws-api.sigmacomputing.com/v2"


@dataclass
class SigmaSourceReport(StaleEntityRemovalSourceReport):
    number_of_workspaces: int = 0
    non_accessible_workspaces_count: int = 0
    shared_entities_count: int = 0
    number_of_datasets: int = 0
    number_of_workbooks: int = 0
    number_of_files_metadata: Dict[str, int] = field(default_factory=dict)

    def report_number_of_workspaces(self, number_of_workspaces: int) -> None:
        self.number_of_workspaces = number_of_workspaces


class PlatformDetail(PlatformInstanceConfigMixin, EnvConfigMixin):
    data_source_platform: str = pydantic.Field(
        description="A chart's data sources platform name.",
    )


class SigmaSourceConfig(
    StatefulIngestionConfigBase, PlatformInstanceConfigMixin, EnvConfigMixin
):
    api_url: str = pydantic.Field(
        default=Constant.DEFAULT_API_URL, description="Sigma API hosted URL."
    )
    client_id: str = pydantic.Field(description="Sigma Client ID")
    client_secret: str = pydantic.Field(description="Sigma Client Secret")
    # Sigma workspace identifier
    workspace_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter Sigma workspaces in ingestion."
        "Mention 'My documents' if personal entities also need to ingest.",
    )
    ingest_owner: Optional[bool] = pydantic.Field(
        default=True,
        description="Ingest Owner from source. This will override Owner info entered from UI.",
    )
    ingest_shared_entities: Optional[bool] = pydantic.Field(
        default=False,
        description="Whether to ingest the shared entities or not.",
    )
    extract_lineage: Optional[bool] = pydantic.Field(
        default=True,
        description="Whether to extract lineage of workbook's elements and datasets or not.",
    )
    workbook_lineage_pattern: AllowDenyPattern = pydantic.Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns to filter workbook's elements and datasets lineage in ingestion."
        "Requires extract_lineage to be enabled.",
    )
    chart_sources_platform_mapping: Dict[str, PlatformDetail] = pydantic.Field(
        default={},
        description="A mapping of the sigma workspace/workbook/chart folder path to all chart's data sources platform details present inside that folder path.",
    )
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = pydantic.Field(
        default=None, description="Sigma Stateful Ingestion Config."
    )



import itertools
import logging
from typing import IO, Dict, List, Type, Union

import jsonlines as jsl
import ujson

from datahub.ingestion.source.schema_inference.base import SchemaInferenceBase
from datahub.ingestion.source.schema_inference.object import construct_schema
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    ArrayTypeClass,
    BooleanTypeClass,
    NullTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaField,
    SchemaFieldDataType,
    StringTypeClass,
    UnionTypeClass,
)

_field_type_mapping: Dict[Union[Type, str], Type] = {
    list: ArrayTypeClass,
    bool: BooleanTypeClass,
    type(None): NullTypeClass,
    int: NumberTypeClass,
    float: NumberTypeClass,
    str: StringTypeClass,
    dict: RecordTypeClass,
    "mixed": UnionTypeClass,
}

logger = logging.getLogger(__name__)


class JsonInferrer(SchemaInferenceBase):
    def __init__(self, max_rows: int = 100, format: str = "json"):
        self.max_rows = max_rows
        self.format = format

    def infer_schema(self, file: IO[bytes]) -> List[SchemaField]:
        if self.format == "jsonl":
            file.seek(0)
            reader = jsl.Reader(file)
            datastore = [
                obj
                for obj in itertools.islice(
                    reader.iter(type=dict, skip_invalid=True), self.max_rows
                )
            ]
        else:
            try:
                datastore = ujson.load(file)
            except ujson.JSONDecodeError as e:
                logger.info(f"Got ValueError: {e}. Retry with jsonlines")
                file.seek(0)
                reader = jsl.Reader(file)
                datastore = [obj for obj in reader.iter(type=dict, skip_invalid=True)]

        if not isinstance(datastore, list):
            datastore = [datastore]

        schema = construct_schema(datastore, delimiter=".")
        fields: List[SchemaField] = []

        for schema_field in schema.values():
            mapped_type = _field_type_mapping.get(schema_field["type"], NullTypeClass)

            native_type = schema_field["type"]

            if isinstance(native_type, type):
                native_type = native_type.__name__

            field = SchemaField(
                fieldPath=schema_field["delimited_name"],
                nativeDataType=native_type,
                type=SchemaFieldDataType(type=mapped_type()),
                nullable=schema_field["nullable"],
                recursive=False,
            )
            fields.append(field)

        return fields



import json
import logging
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple

import pyodata
import pyodata.v2.model
import pyodata.v2.service
from authlib.integrations.requests_client import OAuth2Session
from pydantic import Field, SecretStr, validator
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import (
    DEFAULT_ENV,
    DatasetSourceConfigMixin,
    EnvConfigMixin,
)
from datahub.emitter.mce_builder import (
    dataset_urn_to_key,
    make_dashboard_urn,
    make_data_platform_urn,
    make_dataplatform_instance_urn,
    make_dataset_urn_with_platform_instance,
    make_user_urn,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.incremental_lineage_helper import (
    IncrementalLineageConfigMixin,
    auto_incremental_lineage,
)
from datahub.ingestion.api.source import (
    CapabilityReport,
    MetadataWorkUnitProcessor,
    SourceCapability,
    TestableSource,
    TestConnectionReport,
)
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import BIAssetSubTypes, DatasetSubTypes
from datahub.ingestion.source.sac.sac_common import (
    ImportDataModelColumn,
    Resource,
    ResourceModel,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BrowsePathEntryClass,
    BrowsePathsClass,
    BrowsePathsV2Class,
    ChangeAuditStampsClass,
    DashboardInfoClass,
    DataPlatformInstanceClass,
    DatasetLineageTypeClass,
    DatasetPropertiesClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    SchemaFieldClass,
    SchemaFieldDataTypeClass,
    SchemalessClass,
    SchemaMetadataClass,
    StatusClass,
    StringTypeClass,
    SubTypesClass,
    UpstreamClass,
    UpstreamLineageClass,
)
from datahub.utilities import config_clean

logger = logging.getLogger(__name__)


class ConnectionMappingConfig(EnvConfigMixin):
    platform: Optional[str] = Field(
        default=None, description="The platform that this connection mapping belongs to"
    )

    platform_instance: Optional[str] = Field(
        default=None,
        description="The instance of the platform that this connection mapping belongs to",
    )

    env: str = Field(
        default=DEFAULT_ENV,
        description="The environment that this connection mapping belongs to",
    )


class SACSourceConfig(
    StatefulIngestionConfigBase, DatasetSourceConfigMixin, IncrementalLineageConfigMixin
):
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None,
        description="Stateful ingestion related configs",
    )

    tenant_url: str = Field(description="URL of the SAP Analytics Cloud tenant")
    token_url: str = Field(
        description="URL of the OAuth token endpoint of the SAP Analytics Cloud tenant"
    )
    client_id: str = Field(description="Client ID for the OAuth authentication")
    client_secret: SecretStr = Field(
        description="Client secret for the OAuth authentication"
    )

    ingest_stories: bool = Field(
        default=True,
        description="Controls whether Stories should be ingested",
    )

    ingest_applications: bool = Field(
        default=True,
        description="Controls whether Analytic Applications should be ingested",
    )

    ingest_import_data_model_schema_metadata: bool = Field(
        default=True,
        description="Controls whether schema metadata of Import Data Models should be ingested (ingesting schema metadata of Import Data Models significantly increases overall ingestion time)",
    )

    resource_id_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting resource ids that are to be included",
    )

    resource_name_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting resource names that are to be included",
    )

    folder_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting folders that are to be included",
    )

    connection_mapping: Dict[str, ConnectionMappingConfig] = Field(
        default={}, description="Custom mappings for connections"
    )

    query_name_template: Optional[str] = Field(
        default="QUERY/{name}",
        description="Template for generating dataset urns of consumed queries, the placeholder {query} can be used within the template for inserting the name of the query",
    )

    @validator("tenant_url", "token_url")
    def remove_trailing_slash(cls, v):
        return config_clean.remove_trailing_slashes(v)


@dataclass
class SACSourceReport(StaleEntityRemovalSourceReport):
    pass


@platform_name("SAP Analytics Cloud", id="sac")
@config_class(SACSourceConfig)
@support_status(SupportStatus.TESTING)
@capability(SourceCapability.PLATFORM_INSTANCE, "Enabled by default")
@capability(SourceCapability.DESCRIPTIONS, "Enabled by default")
@capability(
    SourceCapability.LINEAGE_COARSE,
    "Enabled by default (only for Live Data Models)",
)
@capability(SourceCapability.DELETION_DETECTION, "Enabled via stateful ingestion")
@capability(
    SourceCapability.SCHEMA_METADATA,
    "Enabled by default (only for Import Data Models)",
)
class SACSource(StatefulIngestionSourceBase, TestableSource):
    config: SACSourceConfig
    report: SACSourceReport
    platform = "sac"

    session: OAuth2Session
    client: pyodata.Client

    ingested_dataset_entities: Set[str] = set()
    ingested_upstream_dataset_keys: Set[str] = set()

    def __init__(self, config: SACSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.config = config
        self.report = SACSourceReport()

        self.session, self.client = SACSource.get_sac_connection(self.config)

    def close(self) -> None:
        self.session.close()
        super().close()

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "SACSource":
        config = SACSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    @staticmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        test_report = TestConnectionReport()

        try:
            config = SACSourceConfig.parse_obj(config_dict)

            # when creating the pyodata.Client, the metadata is automatically parsed and validated
            session, _ = SACSource.get_sac_connection(config)

            # test the Data Import Service separately here, because it requires specific properties when configuring the OAuth client
            response = session.get(url=f"{config.tenant_url}/api/v1/dataimport/models")
            response.raise_for_status()

            session.close()

            test_report.basic_connectivity = CapabilityReport(capable=True)
        except Exception as e:
            test_report.basic_connectivity = CapabilityReport(
                capable=False, failure_reason=f"{e}"
            )

        return test_report

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            partial(
                auto_incremental_lineage,
                self.config.incremental_lineage,
            ),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        if self.config.ingest_stories or self.config.ingest_applications:
            resources = self.get_resources()

            for resource in resources:
                datasets = []

                for resource_model in resource.resource_models:
                    dataset_urn = make_dataset_urn_with_platform_instance(
                        platform=self.platform,
                        name=f"{resource_model.namespace}:{resource_model.model_id}",
                        platform_instance=self.config.platform_instance,
                        env=self.config.env,
                    )

                    if dataset_urn not in datasets:
                        datasets.append(dataset_urn)

                    if dataset_urn in self.ingested_dataset_entities:
                        continue

                    self.ingested_dataset_entities.add(dataset_urn)

                    yield from self.get_model_workunits(dataset_urn, resource_model)

                yield from self.get_resource_workunits(resource, datasets)

    def get_report(self) -> SACSourceReport:
        return self.report

    def get_resource_workunits(
        self, resource: Resource, datasets: List[str]
    ) -> Iterable[MetadataWorkUnit]:
        dashboard_urn = make_dashboard_urn(
            platform=self.platform,
            name=resource.resource_id,
            platform_instance=self.config.platform_instance,
        )

        if resource.ancestor_path:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dashboard_urn,
                aspect=BrowsePathsClass(
                    paths=[
                        f"/{self.platform}/{resource.ancestor_path}",
                    ],
                ),
            )

            yield mcp.as_workunit()

            mcp = MetadataChangeProposalWrapper(
                entityUrn=dashboard_urn,
                aspect=BrowsePathsV2Class(
                    path=[
                        BrowsePathEntryClass(id=folder_name)
                        for folder_name in resource.ancestor_path.split("/")
                    ],
                ),
            )

            yield mcp.as_workunit()

        if self.config.platform_instance is not None:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dashboard_urn,
                aspect=DataPlatformInstanceClass(
                    platform=make_data_platform_urn(self.platform),
                    instance=make_dataplatform_instance_urn(
                        self.platform, self.config.platform_instance
                    ),
                ),
            )

            yield mcp.as_workunit()

        mcp = MetadataChangeProposalWrapper(
            entityUrn=dashboard_urn,
            aspect=DashboardInfoClass(
                title=resource.name,
                description=resource.description
                if resource.description is not None
                else "",
                lastModified=ChangeAuditStampsClass(
                    created=AuditStampClass(
                        time=round(resource.created_time.timestamp() * 1000),
                        actor=(
                            make_user_urn(resource.created_by)
                            if resource.created_by
                            else "urn:li:corpuser:unknown"
                        ),
                    ),
                    lastModified=AuditStampClass(
                        time=round(resource.modified_time.timestamp() * 1000),
                        actor=(
                            make_user_urn(resource.modified_by)
                            if resource.modified_by
                            else "urn:li:corpuser:unknown"
                        ),
                    ),
                ),
                customProperties={
                    "resourceType": resource.resource_type,
                    "resourceSubtype": resource.resource_subtype,
                    "storyId": resource.story_id,
                    "isMobile": str(resource.is_mobile),
                },
                datasets=sorted(datasets) if datasets else None,
                externalUrl=f"{self.config.tenant_url}{resource.open_url}",
            ),
        )

        yield mcp.as_workunit()

        type_name: Optional[str] = None
        if resource.resource_subtype == "":
            type_name = BIAssetSubTypes.SAC_STORY
        elif resource.resource_subtype == "APPLICATION":
            type_name = BIAssetSubTypes.SAC_APPLICATION

        if type_name:
            mcp = MetadataChangeProposalWrapper(
                entityUrn=dashboard_urn,
                aspect=SubTypesClass(
                    typeNames=[type_name],
                ),
            )

            yield mcp.as_workunit()

    def get_model_workunits(
        self, dataset_urn: str, model: ResourceModel
    ) -> Iterable[MetadataWorkUnit]:
        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=DatasetPropertiesClass(
                name=model.name,
                description=model.description,
                customProperties={
                    "namespace": model.namespace,
                    "modelId": model.model_id,
                    "isImport": "true" if model.is_import else "false",
                },
                externalUrl=f"{self.config.tenant_url}/sap/fpa/ui/tenants/3c44c#view_id=model;model_id={model.namespace}:{model.model_id}",
            ),
        )

        yield mcp.as_workunit()

        if model.is_import and self.config.ingest_import_data_model_schema_metadata:
            primary_fields: List[str] = []
            schema_fields: List[SchemaFieldClass] = []

            columns = self.get_import_data_model_columns(model_id=model.model_id)
            for column in columns:

                schema_field = SchemaFieldClass(
                    fieldPath=column.name,
                    type=self.get_schema_field_data_type(column),
                    nativeDataType=self.get_schema_field_native_data_type(column),
                    description=column.description,
                    isPartOfKey=column.is_key,
                )

                schema_fields.append(schema_field)

                if column.is_key:
                    primary_fields.append(column.name)

            mcp = MetadataChangeProposalWrapper(
                entityUrn=dataset_urn,
                aspect=SchemaMetadataClass(
                    schemaName=model.model_id,
                    platform=make_data_platform_urn(self.platform),
                    version=0,
                    hash="",
                    platformSchema=SchemalessClass(),
                    fields=schema_fields,
                    primaryKeys=primary_fields,
                ),
            )

            yield mcp.as_workunit()

        if model.system_type in ("BW", "HANA") and model.external_id is not None:
            upstream_dataset_name: Optional[str] = None

            if model.system_type == "BW" and model.external_id.startswith(
                "query:"
            ):  # query:[][][query]
                query = model.external_id[11:-1]
                upstream_dataset_name = self.get_query_name(query)
            elif model.system_type == "HANA" and model.external_id.startswith(
                "view:"
            ):  # view:[schema][schema.namespace][view]
                schema, namespace_with_schema, view = model.external_id.split("][", 2)
                schema = schema[6:]
                namespace: Optional[str] = None
                if len(schema) < len(namespace_with_schema):
                    namespace = namespace_with_schema[len(f"{schema}.") :]
                view = view[:-1]
                upstream_dataset_name = self.get_view_name(schema, namespace, view)

            if upstream_dataset_name is not None:
                if model.connection_id in self.config.connection_mapping:
                    connection = self.config.connection_mapping[model.connection_id]
                    platform = (
                        connection.platform
                        if connection.platform
                        else model.system_type.lower()
                    )
                    platform_instance = connection.platform_instance
                    env = connection.env
                else:
                    platform = model.system_type.lower()
                    platform_instance = model.connection_id
                    env = DEFAULT_ENV

                    logger.info(
                        f"No connection mapping found for connection with id {model.connection_id}, connection id will be used as platform instance"
                    )

                upstream_dataset_urn = make_dataset_urn_with_platform_instance(
                    platform=platform,
                    name=upstream_dataset_name,
                    platform_instance=platform_instance,
                    env=env,
                )

                if upstream_dataset_urn not in self.ingested_upstream_dataset_keys:
                    mcp = MetadataChangeProposalWrapper(
                        entityUrn=upstream_dataset_urn,
                        aspect=dataset_urn_to_key(upstream_dataset_urn),
                    )

                    yield mcp.as_workunit(is_primary_source=False)

                    self.ingested_upstream_dataset_keys.add(upstream_dataset_urn)

                mcp = MetadataChangeProposalWrapper(
                    entityUrn=dataset_urn,
                    aspect=UpstreamLineageClass(
                        upstreams=[
                            UpstreamClass(
                                dataset=upstream_dataset_urn,
                                type=DatasetLineageTypeClass.COPY,
                            ),
                        ],
                    ),
                )

                yield mcp.as_workunit()
            else:
                self.report.report_warning(
                    "unknown-upstream-dataset",
                    f"Unknown upstream dataset for model with id {model.namespace}:{model.model_id} and external id {model.external_id}",
                )
        elif model.system_type is not None:
            self.report.report_warning(
                "unknown-system-type",
                f"Unknown system type {model.system_type} for model with id {model.namespace}:{model.model_id} and external id {model.external_id}",
            )

        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=StatusClass(
                removed=False,
            ),
        )

        yield mcp.as_workunit()

        if model.external_id and model.connection_id and model.system_type:
            type_name = DatasetSubTypes.SAC_LIVE_DATA_MODEL
        elif model.is_import:
            type_name = DatasetSubTypes.SAC_IMPORT_DATA_MODEL
        else:
            type_name = DatasetSubTypes.SAC_MODEL

        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=SubTypesClass(
                typeNames=[type_name],
            ),
        )

        yield mcp.as_workunit()

        mcp = MetadataChangeProposalWrapper(
            entityUrn=dataset_urn,
            aspect=DataPlatformInstanceClass(
                platform=make_data_platform_urn(self.platform),
                instance=self.config.platform_instance,
            ),
        )

        yield mcp.as_workunit()

    @staticmethod
    def get_sac_connection(
        config: SACSourceConfig,
    ) -> Tuple[OAuth2Session, pyodata.Client]:
        session = OAuth2Session(
            client_id=config.client_id,
            client_secret=config.client_secret.get_secret_value(),
            token_endpoint=config.token_url,
            token_endpoint_auth_method="client_secret_post",
            grant_type="client_credentials",
        )

        retries = 3
        backoff_factor = 10

        # The Resources and Data Import Service APIs of SAP Analytics Cloud can be somewhat unstable, occasionally
        # returning HTTP errors for some requests, even though the APIs are generally operational. Therefore, we must
        # retry these requests to increase the likelihood that the ingestion is successful. For the same reason we
        # should also retry requests that receive a 401 HTTP status; however, this status also legitimately indicates
        # that the provided OAuth credentials are invalid or that the OAuth client does not have the correct
        # permissions assigned, therefore requests that receive a 401 HTTP status must not be retried.
        status_forcelist = (400, 500, 503)

        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )

        adapter = HTTPAdapter(max_retries=retry)
        session.mount("http://", adapter)
        session.mount("https://", adapter)

        session.register_compliance_hook(
            "protected_request", _add_sap_sac_custom_auth_header
        )
        session.fetch_token()

        client = pyodata.Client(
            url=f"{config.tenant_url}/api/v1",
            connection=session,
            config=pyodata.v2.model.Config(retain_null=True),
        )

        return session, client

    def get_resources(self) -> Iterable[Resource]:
        import_data_model_ids = self.get_import_data_model_ids()

        filter = "isTemplate eq 0 and isSample eq 0 and isPublic eq 1"
        if self.config.ingest_stories and self.config.ingest_applications:
            filter += " and ((resourceType eq 'STORY' and resourceSubtype eq '') or (resourceType eq 'STORY' and resourceSubtype eq 'APPLICATION'))"
        elif self.config.ingest_stories and not self.config.ingest_applications:
            filter += " and resourceType eq 'STORY' and resourceSubtype eq ''"
        elif not self.config.ingest_stories and self.config.ingest_applications:
            filter += (
                " and resourceType eq 'STORY' and resourceSubtype eq 'APPLICATION'"
            )

        select = "resourceId,resourceType,resourceSubtype,storyId,name,description,createdTime,createdBy,modifiedBy,modifiedTime,openURL,ancestorPath,isMobile"

        entities: pyodata.v2.service.ListWithTotalCount = (
            self.client.entity_sets.Resources.get_entities()
            .custom("$format", "json")
            .filter(filter)
            .select(select)
            .execute()
        )
        entity: pyodata.v2.service.EntityProxy
        for entity in entities:
            resource_id: str = entity.resourceId
            name: str = (
                entity.name.strip() if entity.name is not None else entity.resourceId
            )

            if not self.config.resource_id_pattern.allowed(
                resource_id
            ) or not self.config.resource_name_pattern.allowed(name):
                continue

            ancestor_path: Optional[str] = None

            try:
                ancestors = json.loads(entity.ancestorPath)
                ancestor_path = "/".join(
                    ancestor.replace("/", "%2F") for ancestor in ancestors
                )
            except json.JSONDecodeError:
                pass

            if ancestor_path and not self.config.folder_pattern.allowed(ancestor_path):
                continue

            resource_models: Set[ResourceModel] = set()

            select = "modelId,name,description,externalId,connectionId,systemType"

            nav_entities: pyodata.v2.service.EntitySetProxy = (
                entity.nav("resourceModels")
                .get_entities()
                .custom("$format", "json")
                .select(select)
                .execute()
            )
            nav_entity: pyodata.v2.service.EntityProxy
            for nav_entity in nav_entities:
                # the model id can have a different structure, commonly all model ids have a namespace (the part before the colon) and the model id itself
                # t.4.sap.fpa.services.userFriendlyPerfLog:ACTIVITY_LOG is a builtin model without a possiblity to get more metadata about the model
                # t.4.YV67EM4QBRU035A7TVKERZ786N:YV67EM4QBRU035A7TVKERZ786N is a model id where the model id itself also appears as part of the namespace
                # t.4:C76tt2j402o1e69wnvrwfcl79c is a model id without the model id itself as part of the namespace
                model_id: str = nav_entity.modelId
                namespace, _, model_id = model_id.partition(":")

                resource_models.add(
                    ResourceModel(
                        namespace=namespace,
                        model_id=model_id,
                        name=nav_entity.name.strip()
                        if nav_entity.name is not None
                        else f"{namespace}:{model_id}",
                        description=nav_entity.description.strip()
                        if nav_entity.description is not None
                        else None,
                        system_type=nav_entity.systemType,  # BW or HANA
                        connection_id=nav_entity.connectionId,
                        external_id=nav_entity.externalId,  # query:[][][query] or view:[schema][schema.namespace][view]
                        is_import=model_id in import_data_model_ids,
                    )
                )

            created_by: Optional[str] = entity.createdBy
            if created_by in ("SYSTEM", "$DELETED_USER$"):
                created_by = None

            modified_by: Optional[str] = entity.modifiedBy
            if modified_by in ("SYSTEM", "$DELETED_USER$"):
                modified_by = None

            yield Resource(
                resource_id=resource_id,
                resource_type=entity.resourceType,
                resource_subtype=entity.resourceSubtype,
                story_id=entity.storyId,
                name=name,
                description=entity.description.strip()
                if entity.description is not None
                else None,
                created_time=entity.createdTime,
                created_by=created_by,
                modified_time=entity.modifiedTime,
                modified_by=modified_by,
                open_url=entity.openURL,
                ancestor_path=ancestor_path,
                is_mobile=entity.isMobile,
                resource_models=frozenset(resource_models),
            )

    def get_import_data_model_ids(self) -> Set[str]:
        response = self.session.get(
            url=f"{self.config.tenant_url}/api/v1/dataimport/models"
        )
        response.raise_for_status()

        import_data_model_ids = set(
            model["modelID"] for model in response.json()["models"]
        )
        return import_data_model_ids

    def get_import_data_model_columns(
        self, model_id: str
    ) -> List[ImportDataModelColumn]:
        response = self.session.get(
            url=f"{self.config.tenant_url}/api/v1/dataimport/models/{model_id}/metadata"
        )
        response.raise_for_status()

        model_metadata = response.json()

        columns: List[ImportDataModelColumn] = []
        for column in model_metadata["factData"]["columns"]:
            columns.append(
                ImportDataModelColumn(
                    name=column["columnName"].strip(),
                    description=(
                        column["descriptionName"].strip()
                        if column.get("descriptionName") is not None
                        else None
                    ),
                    property_type=column["propertyType"],
                    data_type=column["columnDataType"],
                    max_length=column.get("maxLength"),
                    precision=column.get("precision"),
                    scale=column.get("scale"),
                    is_key=column["isKey"],
                )
            )

        return columns

    def get_query_name(self, query: str) -> str:
        if not self.config.query_name_template:
            return query

        query_name = self.config.query_name_template
        query_name = query_name.replace("{name}", query)

        return query_name

    def get_view_name(self, schema: str, namespace: Optional[str], view: str) -> str:
        if namespace:
            return f"{schema}.{namespace}::{view}"

        return f"{schema}.{view}"

    def get_schema_field_data_type(
        self, column: ImportDataModelColumn
    ) -> SchemaFieldDataTypeClass:
        if column.property_type == "DATE":
            return SchemaFieldDataTypeClass(type=DateTypeClass())
        else:
            if column.data_type == "string":
                return SchemaFieldDataTypeClass(type=StringTypeClass())
            elif column.data_type in ("decimal", "int32"):
                return SchemaFieldDataTypeClass(type=NumberTypeClass())
            else:
                self.report.report_warning(
                    "unknown-data-type",
                    f"Unknown data type {column.data_type} found",
                )

                return SchemaFieldDataTypeClass(type=NullTypeClass())

    def get_schema_field_native_data_type(self, column: ImportDataModelColumn) -> str:
        native_data_type = column.data_type
        if column.data_type == "decimal":
            native_data_type = f"{column.data_type}({column.precision}, {column.scale})"
        elif column.data_type == "int32":
            native_data_type = f"{column.data_type}({column.precision})"
        elif column.max_length is not None:
            native_data_type = f"{column.data_type}({column.max_length})"

        return native_data_type


def _add_sap_sac_custom_auth_header(
    url: str, headers: Dict[str, str], body: Any
) -> Tuple[str, Dict[str, str], Any]:
    headers["x-sap-sac-custom-auth"] = "true"
    return url, headers, body



import logging
import time
from typing import Any, Dict, Iterator, Optional

from pydantic import BaseModel, Field

from datahub.configuration.common import ConfigModel
from datahub.ingestion.api.source import SourceReport
from datahub.ingestion.graph.client import DataHubGraph

logger = logging.getLogger(__name__)

DATAHUB_EXECUTION_REQUEST_ENTITY_NAME = "dataHubExecutionRequest"
DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME = "dataHubExecutionRequestKey"
DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME = "dataHubExecutionRequestInput"
DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME = "dataHubExecutionRequestResult"


class DatahubExecutionRequestCleanupConfig(ConfigModel):
    keep_history_min_count: int = Field(
        10,
        description="Minimum number of execution requests to keep, per ingestion source",
    )

    keep_history_max_count: int = Field(
        1000,
        description="Maximum number of execution requests to keep, per ingestion source",
    )

    keep_history_max_days: int = Field(
        30,
        description="Maximum number of days to keep execution requests for, per ingestion source",
    )

    batch_read_size: int = Field(
        100,
        description="Number of records per read operation",
    )

    enabled: bool = Field(
        True,
        description="Global switch for this cleanup task",
    )

    def keep_history_max_milliseconds(self):
        return self.keep_history_max_days * 24 * 3600 * 1000


class DatahubExecutionRequestCleanupReport(SourceReport):
    execution_request_cleanup_records_read: int = 0
    execution_request_cleanup_records_preserved: int = 0
    execution_request_cleanup_records_deleted: int = 0
    execution_request_cleanup_read_errors: int = 0
    execution_request_cleanup_delete_errors: int = 0


class CleanupRecord(BaseModel):
    urn: str
    request_id: str
    status: str
    ingestion_source: str
    requested_at: int


class DatahubExecutionRequestCleanup:
    def __init__(
        self,
        graph: DataHubGraph,
        report: DatahubExecutionRequestCleanupReport,
        config: Optional[DatahubExecutionRequestCleanupConfig] = None,
    ) -> None:

        self.graph = graph
        self.report = report
        self.instance_id = int(time.time())

        if config is not None:
            self.config = config
        else:
            self.config = DatahubExecutionRequestCleanupConfig()

    def _to_cleanup_record(self, entry: Dict) -> CleanupRecord:
        input_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME, {})
            .get("value", {})
        )
        result_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME, {})
            .get("value", {})
        )
        key_aspect = (
            entry.get("aspects", {})
            .get(DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME, {})
            .get("value", {})
        )
        return CleanupRecord(
            urn=entry.get("urn"),
            request_id=key_aspect.get("id"),
            requested_at=input_aspect.get("requestedAt", 0),
            status=result_aspect.get("status", "PENDING"),
            ingestion_source=input_aspect.get("source", {}).get("ingestionSource", ""),
        )

    def _scroll_execution_requests(
        self, overrides: Dict[str, Any] = {}
    ) -> Iterator[CleanupRecord]:
        headers: Dict[str, Any] = {
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        params = {
            "aspectNames": [
                DATAHUB_EXECUTION_REQUEST_KEY_ASPECT_NAME,
                DATAHUB_EXECUTION_REQUEST_INPUT_ASPECT_NAME,
                DATAHUB_EXECUTION_REQUEST_RESULT_ASPECT_NAME,
            ],
            "count": str(self.config.batch_read_size),
            "sort": "requestTimeMs",
            "sortOrder": "DESCENDING",
            "systemMetadata": "false",
            "skipCache": "true",
        }
        params.update(overrides)

        while True:
            try:
                url = f"{self.graph.config.server}/openapi/v2/entity/{DATAHUB_EXECUTION_REQUEST_ENTITY_NAME}"
                response = self.graph._session.get(url, headers=headers, params=params)
                response.raise_for_status()
                document = response.json()

                entries = document.get("results", [])
                for entry in entries:
                    yield self._to_cleanup_record(entry)

                if "scrollId" not in document:
                    break
                params["scrollId"] = document["scrollId"]
            except Exception as e:
                logger.error(
                    f"ergc({self.instance_id}): failed to fetch next batch of execution requests: {e}"
                )
                self.report.execution_request_cleanup_read_errors += 1

    def _scroll_garbage_records(self):
        state: Dict[str, Dict] = {}

        now_ms = int(time.time()) * 1000
        running_guard_timeout = now_ms - 30 * 24 * 3600 * 1000

        for entry in self._scroll_execution_requests():
            self.report.execution_request_cleanup_records_read += 1
            key = entry.ingestion_source

            # Always delete corrupted records
            if not key:
                logger.warning(
                    f"ergc({self.instance_id}): will delete corrupted entry with missing source key: {entry}"
                )
                yield entry
                continue

            if key not in state:
                state[key] = {}
                state[key]["cutoffTimestamp"] = (
                    entry.requested_at - self.config.keep_history_max_milliseconds()
                )

            state[key]["count"] = state[key].get("count", 0) + 1

            # Do not delete if number of requests is below minimum
            if state[key]["count"] < self.config.keep_history_min_count:
                self.report.execution_request_cleanup_records_preserved += 1
                continue

            # Do not delete if number of requests do not exceed allowed maximum,
            # or the cutoff date.
            if (state[key]["count"] < self.config.keep_history_max_count) and (
                entry.requested_at > state[key]["cutoffTimestamp"]
            ):
                self.report.execution_request_cleanup_records_preserved += 1
                continue

            # Do not delete if status is RUNNING or PENDING and created within last month. If the record is >month old and it did not
            # transition to a final state within that timeframe, it likely has no value.
            if entry.requested_at > running_guard_timeout and entry.status in [
                "RUNNING",
                "PENDING",
            ]:
                self.report.execution_request_cleanup_records_preserved += 1
                continue

            # Otherwise delete current record
            logger.info(
                (
                    f"ergc({self.instance_id}): going to delete {entry.request_id} in source {key}; "
                    f"source count: {state[key]['count']}; "
                    f"source cutoff: {state[key]['cutoffTimestamp']}; "
                    f"record timestamp: {entry.requested_at}."
                )
            )
            self.report.execution_request_cleanup_records_deleted += 1
            yield entry

    def _delete_entry(self, entry: CleanupRecord) -> None:
        try:
            logger.info(
                f"ergc({self.instance_id}): going to delete ExecutionRequest {entry.request_id}"
            )
            self.graph.delete_entity(entry.urn, True)
        except Exception as e:
            self.report.execution_request_cleanup_delete_errors += 1
            logger.error(
                f"ergc({self.instance_id}): failed to delete ExecutionRequest {entry.request_id}: {e}"
            )

    def run(self) -> None:
        if not self.config.enabled:
            logger.info(
                f"ergc({self.instance_id}): ExecutionRequest cleaner is disabled."
            )
            return

        logger.info(
            (
                f"ergc({self.instance_id}): Starting cleanup of ExecutionRequest records; "
                f"max days: {self.config.keep_history_max_days}, "
                f"min records: {self.config.keep_history_min_count}, "
                f"max records: {self.config.keep_history_max_count}."
            )
        )

        for entry in self._scroll_garbage_records():
            self._delete_entry(entry)

        logger.info(
            f"ergc({self.instance_id}): Finished cleanup of ExecutionRequest records."
        )
